# -*- coding: utf-8 -*-
# vispy: testskip
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
# Author:   Per Rosengren
# Date:     18/03/2014
# Abstract: Unstructured2D canvas example
# Keywords: unstructured delaunay colormap
# Require: scipy
# ----------------------------------------------------------------------------
"""Unstructured2D canvas example.

Takes unstructured 2D locations, with corresponding 1 or 2 dimensional
scalar "values". Plots the values looked up from colormaps and
interpolated between the locations.
"""
import sys
import numpy as np
import scipy.spatial

from vispy import gloo, app
from vispy.util.transforms import ortho


class Canvas(app.Canvas):

    def __init__(self,
                 x=None, y=None, u=None, v=None,
                 colormap=None, data_lim=None,
                 dir_x_right=True, dir_y_top=True,
                 **kwargs):
        app.Canvas.__init__(self, **kwargs)
        self.create_shader(colormap)
        self.create_mesh(x, y, u, v)
        self.program.bind(self.vbo)
        if data_lim is not None:
            self._data_lim = data_lim
        else:
            self._data_lim = [[x.min(), x.max()], [y.min(), y.max()]]
        self._dir_x_right = dir_x_right
        self._dir_y_top = dir_y_top

        self.activate_zoom()

        self.show()

    def create_shader(self, colormap):
        if len(colormap.shape) == 2:
            args = dict(
                n_dims="1",
                tex_t="float",
                texture2D_arg="vec2(v_texcoord, 0.)")
        else:
            args = dict(
                n_dims="2",
                tex_t="vec2",
                texture2D_arg="v_texcoord")
        vertex = """
            uniform mat4 projection;
            uniform sampler2D texture;

            attribute vec2 position;
            attribute {tex_t} texcoord;

            varying {tex_t} v_texcoord;
            void main()
            {{
                gl_Position = projection * vec4(position, 0.0, 1.0);
                v_texcoord = texcoord;
            }}
        """.format(**args)

        fragment = """
            uniform sampler2D texture;
            varying {tex_t} v_texcoord;
            void main()
            {{
                gl_FragColor = texture2D(texture, {texture2D_arg});
            }}
        """.format(**args)

        self.program = gloo.Program(vertex, fragment)
        if len(colormap.shape) == 2:
            self.program['texture'] = np.ascontiguousarray(
                colormap[None, :, :])
        else:
            self.program['texture'] = colormap
        self.program['texture'].interpolation = 'linear'
        self.projection = np.eye(4, dtype=np.float32)

    def create_mesh(self, x, y, u, v):
        tri = scipy.spatial.Delaunay(np.column_stack([x, y]))
        edges = tri.simplices.astype(np.uint32)
        uv = []
        for c in [u, v]:
            if c is not None:
                c = c.astype('f4')
                c = .5 + .5 * c / np.abs(c).max()
                uv.append(c)
        data = np.column_stack(
            [
                x.astype('f4'),
                y.astype('f4')
            ] + uv
        ).view(dtype=[('position', 'f4', (2,)),
                      ('texcoord', 'f4', (2 if v is not None else 1,)),
                      ])
        self.vbo = gloo.VertexBuffer(data)
        self.index = gloo.IndexBuffer(edges)

        gloo.set_state(blend=True, clear_color='white',
                       blend_func=('src_alpha', 'one_minus_src_alpha'))

    def on_draw(self, event):
        gloo.clear()
        self.program.draw('triangles', self.index)

    def on_resize(self, event):
        self.activate_zoom()

    def activate_zoom(self):
        width, heigh = self.size
        gloo.set_viewport(0, 0, *self.physical_size)
        data_width = self._data_lim[0][1] - self._data_lim[0][0]
        data_height = self._data_lim[1][1] - self._data_lim[1][0]
        data_aspect = data_width / float(data_height)
        frame_aspect = width / float(height)
        if frame_aspect >= data_aspect:
            padding = (frame_aspect * data_height - data_width) / 2.
            frame_lim = [
                [self._data_lim[0][0] - padding,
                 self._data_lim[0][1] + padding],
                [self._data_lim[1][0],
                 self._data_lim[1][1]]]
        else:
            padding = (data_width / frame_aspect - data_height) / 2.
            frame_lim = [
                [self._data_lim[0][0],
                 self._data_lim[0][1]],
                [self._data_lim[1][0] - padding,
                 self._data_lim[1][1] + padding]]
        args_ortho = frame_lim[0][::(1 if self._dir_x_right else -1)]
        args_ortho += frame_lim[1][::(1 if self._dir_y_top else -1)]
        args_ortho += -1000, 1000
        self.projection = ortho(*args_ortho)
        self.program['projection'] = self.projection


def create_colormap2d_hsv(size=512):
    import matplotlib.colors
    import math
    u, v = np.meshgrid(np.linspace(-1, 1, size), np.linspace(-1, 1, size))
    hsv = np.ones((size, size, 3), dtype=np.float32)
    hsv[:, :, 0] = (np.arctan2(u, v) / (2 * math.pi) + .5)
    hsv[:, :, 1] = np.minimum(1., np.sqrt(u ** 2 + v ** 2))
    rgb = matplotlib.colors.hsv_to_rgb(hsv)
    return rgb


def create_colormap2d_4dirs(size=512):
    rgb = np.ones((size, size, 3), dtype=np.float32)
    hs = size // 2
    u, v = np.meshgrid(np.linspace(1, 0, hs), np.linspace(1, 0, hs))
    rgb[:hs, :hs, 0] = 1.
    rgb[:hs, :hs, 1] = 1. - v + u / 2.
    rgb[:hs, :hs, 2] = 1. - np.maximum(u, v)
    u = u[:, ::-1]
    rgb[:hs, hs:, 0] = 1. - u + v
    rgb[:hs, hs:, 1] = 1. - np.maximum(u, v)
    rgb[:hs, hs:, 2] = 1. - v + u
    v = v[::-1, :]
    rgb[hs:, hs:, 0] = 1. - np.maximum(u, v)
    rgb[hs:, hs:, 1] = 1. - u + v
    rgb[hs:, hs:, 2] = 1. - v + u
    u = u[:, ::-1]
    rgb[hs:, :hs, 0] = 1. - v + u / 2.
    rgb[hs:, :hs, 1] = 1.
    rgb[hs:, :hs, 2] = 1. - np.maximum(u, v)
    rgb = np.minimum(1., rgb)
    return rgb


def create_colormap1d_hot(size=512):
    rgb = np.ones((size, 3), dtype=np.float32)
    hs = size // 2
    u = np.linspace(1, 0, hs)
    rgb[:hs, 0] = 1 - u
    rgb[:hs, 1] = 1 - u
    u = u[::-1]
    rgb[hs:, 1] = 1 - u
    rgb[hs:, 2] = 1 - u
    return rgb


if __name__ == '__main__':
    loc = np.random.random_sample(size=(100, 2))
    np.random.shuffle(loc)
    vec = np.empty_like(loc)
    vec[:, 0] = np.cos(loc[:, 0] * 10)
    vec[:, 1] = np.cos(loc[:, 1] * 13)
    width = 500
    height = 500
    c1 = Canvas(title="Unstructured 2D - 2D colormap",
                size=(width, height), position=(0, 40),
                x=loc[:, 0], y=loc[:, 1], u=vec[:, 0], v=vec[:, 1],
                colormap=create_colormap2d_4dirs(size=128),
                keys='interactive')
    c2 = Canvas(title="Unstructured 2D - 1D colormap",
                size=(width, height), position=(width + 20, 40),
                x=loc[:, 0], y=loc[:, 1], u=vec[:, 0],
                colormap=create_colormap1d_hot(size=128),
                keys='interactive')
    if sys.flags.interactive == 0:
        app.run()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
# Author: Nicolas P .Rougier
# Date:   04/03/2014
# -----------------------------------------------------------------------------
"""
Drawing a rotated 3D cube
=========================
"""
import math
import numpy as np

from vispy import app
from vispy.gloo import gl


def checkerboard(grid_num=8, grid_size=32):
    row_even = grid_num // 2 * [0, 1]
    row_odd = grid_num // 2 * [1, 0]
    Z = np.row_stack(grid_num // 2 * (row_even, row_odd)).astype(np.uint8)
    return 255 * Z.repeat(grid_size, axis=0).repeat(grid_size, axis=1)


def rotate(M, angle, x, y, z, point=None):
    angle = math.pi * angle / 180
    c, s = math.cos(angle), math.sin(angle)
    n = math.sqrt(x * x + y * y + z * z)
    x /= n
    y /= n
    z /= n
    cx, cy, cz = (1 - c) * x, (1 - c) * y, (1 - c) * z
    R = np.array([[cx * x + c, cy * x - z * s, cz * x + y * s, 0],
                  [cx * y + z * s, cy * y + c, cz * y - x * s, 0],
                  [cx * z - y * s, cy * z + x * s, cz * z + c, 0],
                  [0, 0, 0, 1]], dtype=M.dtype).T
    M[...] = np.dot(M, R)
    return M


def translate(M, x, y=None, z=None):
    y = x if y is None else y
    z = x if z is None else z
    T = np.array([[1.0, 0.0, 0.0, x],
                  [0.0, 1.0, 0.0, y],
                  [0.0, 0.0, 1.0, z],
                  [0.0, 0.0, 0.0, 1.0]], dtype=M.dtype).T
    M[...] = np.dot(M, T)
    return M


def frustum(left, right, bottom, top, znear, zfar):
    M = np.zeros((4, 4), dtype=np.float32)
    M[0, 0] = +2.0 * znear / (right - left)
    M[2, 0] = (right + left) / (right - left)
    M[1, 1] = +2.0 * znear / (top - bottom)
    M[3, 1] = (top + bottom) / (top - bottom)
    M[2, 2] = -(zfar + znear) / (zfar - znear)
    M[3, 2] = -2.0 * znear * zfar / (zfar - znear)
    M[2, 3] = -1.0
    return M


def perspective(fovy, aspect, znear, zfar):
    h = math.tan(fovy / 360.0 * math.pi) * znear
    w = h * aspect
    return frustum(-w, w, -h, h, znear, zfar)


def makecube():
    """ Generate vertices & indices for a filled cube """

    vtype = [('a_position', np.float32, 3),
             ('a_texcoord', np.float32, 2)]
    itype = np.uint32

    # Vertices positions
    p = np.array([[1, 1, 1], [-1, 1, 1], [-1, -1, 1], [1, -1, 1],
                  [1, -1, -1], [1, 1, -1], [-1, 1, -1], [-1, -1, -1]])

    # Texture coords
    t = np.array([[0, 0], [0, 1], [1, 1], [1, 0]])

    faces_p = [0, 1, 2, 3, 0, 3, 4, 5, 0, 5, 6,
               1, 1, 6, 7, 2, 7, 4, 3, 2, 4, 7, 6, 5]
    faces_t = [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2,
               3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3]

    vertices = np.zeros(24, vtype)
    vertices['a_position'] = p[faces_p]
    vertices['a_texcoord'] = t[faces_t]

    indices = np.resize(
        np.array([0, 1, 2, 0, 2, 3], dtype=itype), 6 * (2 * 3))
    indices += np.repeat(4 * np.arange(6), 6).astype(np.uint32)

    return vertices, indices


cube_vertex = """
uniform mat4 u_model;
uniform mat4 u_view;
uniform mat4 u_projection;
attribute vec3 a_position;
attribute vec2 a_texcoord;
varying vec2 v_texcoord;
void main()
{
    gl_Position = u_projection * u_view * u_model * vec4(a_position,1.0);
    v_texcoord = a_texcoord;
}
"""

cube_fragment = """
uniform sampler2D u_texture;
varying vec2 v_texcoord;
void main()
{
    gl_FragColor = texture2D(u_texture, v_texcoord);
}
"""


class Canvas(app.Canvas):
    def __init__(self):
        app.Canvas.__init__(self, size=(512, 512),
                            title='Rotating cube (GL version)',
                            keys='interactive')

    def on_initialize(self, event):
        # Build & activate cube program
        self.cube = gl.glCreateProgram()
        vertex = gl.glCreateShader(gl.GL_VERTEX_SHADER)
        fragment = gl.glCreateShader(gl.GL_FRAGMENT_SHADER)
        gl.glShaderSource(vertex, cube_vertex)
        gl.glShaderSource(fragment, cube_fragment)
        gl.glCompileShader(vertex)
        gl.glCompileShader(fragment)
        gl.glAttachShader(self.cube, vertex)
        gl.glAttachShader(self.cube, fragment)
        gl.glLinkProgram(self.cube)
        gl.glDetachShader(self.cube, vertex)
        gl.glDetachShader(self.cube, fragment)
        gl.glUseProgram(self.cube)

        # Get data & build cube buffers
        vcube_data, self.icube_data = makecube()
        vcube = gl.glCreateBuffer()
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, vcube)
        gl.glBufferData(gl.GL_ARRAY_BUFFER, vcube_data, gl.GL_STATIC_DRAW)
        icube = gl.glCreateBuffer()
        gl.glBindBuffer(gl.GL_ELEMENT_ARRAY_BUFFER, icube)
        gl.glBufferData(gl.GL_ELEMENT_ARRAY_BUFFER,
                        self.icube_data, gl.GL_STATIC_DRAW)

        # Bind cube attributes
        stride = vcube_data.strides[0]
        offset = 0
        loc = gl.glGetAttribLocation(self.cube, "a_position")
        gl.glEnableVertexAttribArray(loc)
        gl.glVertexAttribPointer(loc, 3, gl.GL_FLOAT, False, stride, offset)

        offset = vcube_data.dtype["a_position"].itemsize
        loc = gl.glGetAttribLocation(self.cube, "a_texcoord")
        gl.glEnableVertexAttribArray(loc)
        gl.glVertexAttribPointer(loc, 2, gl.GL_FLOAT, False, stride, offset)

        # Create & bind cube texture
        crate = checkerboard()
        texture = gl.glCreateTexture()
        gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER,
                           gl.GL_LINEAR)
        gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER,
                           gl.GL_LINEAR)
        gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S,
                           gl.GL_CLAMP_TO_EDGE)
        gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T,
                           gl.GL_CLAMP_TO_EDGE)
        gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_LUMINANCE, gl.GL_LUMINANCE,
                        gl.GL_UNSIGNED_BYTE, crate.shape[:2])
        gl.glTexSubImage2D(gl.GL_TEXTURE_2D, 0, 0, 0, gl.GL_LUMINANCE,
                           gl.GL_UNSIGNED_BYTE, crate)
        loc = gl.glGetUniformLocation(self.cube, "u_texture")
        gl.glUniform1i(loc, texture)
        gl.glBindTexture(gl.GL_TEXTURE_2D, 0)

        # Create & bind cube matrices
        view = np.eye(4, dtype=np.float32)
        model = np.eye(4, dtype=np.float32)
        projection = np.eye(4, dtype=np.float32)
        translate(view, 0, 0, -7)
        self.phi, self.theta = 60, 20
        rotate(model, self.theta, 0, 0, 1)
        rotate(model, self.phi, 0, 1, 0)
        loc = gl.glGetUniformLocation(self.cube, "u_model")
        gl.glUniformMatrix4fv(loc, 1, False, model)
        loc = gl.glGetUniformLocation(self.cube, "u_view")
        gl.glUniformMatrix4fv(loc, 1, False, view)
        loc = gl.glGetUniformLocation(self.cube, "u_projection")
        gl.glUniformMatrix4fv(loc, 1, False, projection)

        # OpenGL initalization
        gl.glClearColor(0.30, 0.30, 0.35, 1.00)
        gl.glEnable(gl.GL_DEPTH_TEST)
        self._resize(*(self.size + self.physical_size))
        self.timer = app.Timer('auto', self.on_timer, start=True)

    def on_draw(self, event):
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glDrawElements(gl.GL_TRIANGLES, self.icube_data.size,
                          gl.GL_UNSIGNED_INT, None)

    def on_resize(self, event):
        self._resize(*(event.size + event.physical_size))

    def _resize(self, width, height, physical_width, physical_height):
        gl.glViewport(0, 0, physical_width, physical_height)
        projection = perspective(35.0, width / float(height), 2.0, 10.0)
        loc = gl.glGetUniformLocation(self.cube, "u_projection")
        gl.glUniformMatrix4fv(loc, 1, False, projection)

    def on_timer(self, event):
        self.theta += .5
        self.phi += .5
        model = np.eye(4, dtype=np.float32)
        rotate(model, self.theta, 0, 0, 1)
        rotate(model, self.phi, 0, 1, 0)
        loc = gl.glGetUniformLocation(self.cube, "u_model")
        gl.glUniformMatrix4fv(loc, 1, False, model)
        self.update()

if __name__ == '__main__':
    c = Canvas()
    c.show()
    app.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vispy: gallery 60

"""
Dynamic planar graph layout.
"""

import numpy as np
from vispy import gloo, app
from vispy.gloo import set_viewport, set_state, clear

vert = """
#version 120

// Uniforms
// ------------------------------------
uniform mat4 u_model;
uniform mat4 u_view;
uniform mat4 u_projection;
uniform float u_antialias;
uniform float u_size;

// Attributes
// ------------------------------------
attribute vec3  a_position;
attribute vec4  a_fg_color;
attribute vec4  a_bg_color;
attribute float a_linewidth;
attribute float a_size;

// Varyings
// ------------------------------------
varying vec4 v_fg_color;
varying vec4 v_bg_color;
varying float v_size;
varying float v_linewidth;
varying float v_antialias;

void main (void) {
    v_size = a_size * u_size;
    v_linewidth = a_linewidth;
    v_antialias = u_antialias;
    v_fg_color  = a_fg_color;
    v_bg_color  = a_bg_color;
    gl_Position = u_projection * u_view * u_model *
        vec4(a_position*u_size,1.0);
    gl_PointSize = v_size + 2.*(v_linewidth + 1.5*v_antialias);
}
"""

frag = """
#version 120

// Constants
// ------------------------------------

// Varyings
// ------------------------------------
varying vec4 v_fg_color;
varying vec4 v_bg_color;
varying float v_size;
varying float v_linewidth;
varying float v_antialias;

// Functions
// ------------------------------------
float marker(vec2 P, float size);


// Main
// ------------------------------------
void main()
{
    float size = v_size +2*(v_linewidth + 1.5*v_antialias);
    float t = v_linewidth/2.0-v_antialias;

    // The marker function needs to be linked with this shader
    float r = marker(gl_PointCoord, size);

    float d = abs(r) - t;
    if( r > (v_linewidth/2.0+v_antialias))
    {
        discard;
    }
    else if( d < 0.0 )
    {
       gl_FragColor = v_fg_color;
    }
    else
    {
        float alpha = d/v_antialias;
        alpha = exp(-alpha*alpha);
        if (r > 0)
            gl_FragColor = vec4(v_fg_color.rgb, alpha*v_fg_color.a);
        else
            gl_FragColor = mix(v_bg_color, v_fg_color, alpha);
    }
}

float marker(vec2 P, float size)
{
    float r = length((P.xy - vec2(0.5,0.5))*size);
    r -= v_size/2.;
    return r;
}
"""

vs = """
attribute vec3 a_position;
attribute vec4 a_fg_color;
attribute vec4 a_bg_color;
attribute float a_size;
attribute float a_linewidth;

void main(){
    gl_Position = vec4(a_position, 1.);
}
"""

fs = """
void main(){
    gl_FragColor = vec4(0., 0., 0., 1.);
}
"""


class Canvas(app.Canvas):

    def __init__(self, **kwargs):
        # Initialize the canvas for real
        app.Canvas.__init__(self, keys='interactive', size=(512, 512),
                            **kwargs)
        ps = self.pixel_scale
        self.position = 50, 50

        n = 100
        ne = 100
        data = np.zeros(n, dtype=[('a_position', np.float32, 3),
                                  ('a_fg_color', np.float32, 4),
                                  ('a_bg_color', np.float32, 4),
                                  ('a_size', np.float32),
                                  ('a_linewidth', np.float32),
                                  ])
        edges = np.random.randint(size=(ne, 2), low=0,
                                  high=n).astype(np.uint32)
        data['a_position'] = np.hstack((.25 * np.random.randn(n, 2),
                                        np.zeros((n, 1))))
        data['a_fg_color'] = 0, 0, 0, 1
        color = np.random.uniform(0.5, 1., (n, 3))
        data['a_bg_color'] = np.hstack((color, np.ones((n, 1))))
        data['a_size'] = np.random.randint(size=n, low=8*ps, high=20*ps)
        data['a_linewidth'] = 1.*ps
        u_antialias = 1

        self.vbo = gloo.VertexBuffer(data)
        self.index = gloo.IndexBuffer(edges)
        self.view = np.eye(4, dtype=np.float32)
        self.model = np.eye(4, dtype=np.float32)
        self.projection = np.eye(4, dtype=np.float32)

        self.program = gloo.Program(vert, frag)
        self.program.bind(self.vbo)
        self.program['u_size'] = 1
        self.program['u_antialias'] = u_antialias
        self.program['u_model'] = self.model
        self.program['u_view'] = self.view
        self.program['u_projection'] = self.projection

        set_viewport(0, 0, *self.physical_size)

        self.program_e = gloo.Program(vs, fs)
        self.program_e.bind(self.vbo)

        set_state(clear_color='white', depth_test=False, blend=True,
                  blend_func=('src_alpha', 'one_minus_src_alpha'))

        self.show()

    def on_resize(self, event):
        set_viewport(0, 0, *event.physical_size)

    def on_draw(self, event):
        clear(color=True, depth=True)
        self.program_e.draw('lines', self.index)
        self.program.draw('points')

if __name__ == '__main__':
    c = Canvas(title="Graph")
    app.run()



# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# vispy: gallery 2
"""
Zoom Sensitivity
================

For testing zoom sensitivity on various platforms

"""

import numpy as np
import vispy.scene
from vispy.scene import visuals

canvas = vispy.scene.SceneCanvas(keys='interactive', show=True)
vb = canvas.central_widget.add_view()
vb.camera = 'panzoom'
vb.camera.rect = (-10, -10, 20, 20)

centers = np.random.normal(size=(50, 2))
pos = np.random.normal(size=(100000, 2), scale=0.2)
indexes = np.random.normal(size=100000, loc=centers.shape[0]/2., 
                           scale=centers.shape[0]/3.)
indexes = np.clip(indexes, 0, centers.shape[0]-1).astype(int)
scales = 10**(np.linspace(-2, 0.5, centers.shape[0]))[indexes][:, np.newaxis]
pos *= scales
pos += centers[indexes]

scatter = visuals.Markers()
scatter.set_gl_state('translucent', depth_test=False)
scatter.set_data(pos, edge_width=0, face_color=(1, 1, 1, 0.3), size=5)
vb.add(scatter)


@canvas.connect
def on_key_press(ev):
    if ev.key.name in '+=':
        vb.camera.zoom_factor *= 1.1
    elif ev.key.name == '-':
        vb.camera.zoom_factor /= 1.1
    print("Zoom factor: %0.4f" % vb.camera.zoom_factor)


if __name__ == '__main__':
    import sys
    if sys.flags.interactive != 1:
        vispy.app.run()



# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# vispy: gallery 1
"""
Multi-view Volume Plot
======================

Plot various views of a structural MRI.
"""

import numpy as np

from vispy import io, plot as vp

fig = vp.Fig(bgcolor='k', size=(800, 800), show=False)

vol_data = np.load(io.load_data_file('brain/mri.npz'))['data']
vol_data = np.flipud(np.rollaxis(vol_data, 1))
vol_data = vol_data.astype(np.float32)

clim = [32, 192]
texture_format = "auto"  # None for CPUScaled, "auto" for GPUScaled

vol_pw = fig[0, 0]
v = vol_pw.volume(vol_data, clim=clim, texture_format=texture_format)
vol_pw.view.camera.elevation = 30
vol_pw.view.camera.azimuth = 30
vol_pw.view.camera.scale_factor /= 1.5

shape = vol_data.shape
fig[1, 0].image(vol_data[:, :, shape[2] // 2], cmap='grays', clim=clim,
                fg_color=(0.5, 0.5, 0.5, 1), texture_format=texture_format)
fig[0, 1].image(vol_data[:, shape[1] // 2, :], cmap='grays', clim=clim,
                fg_color=(0.5, 0.5, 0.5, 1), texture_format=texture_format)
fig[1, 1].image(vol_data[shape[0] // 2, :, :].T, cmap='grays', clim=clim,
                fg_color=(0.5, 0.5, 0.5, 1), texture_format=texture_format)

if __name__ == '__main__':
    fig.show(run=True)



# -*- coding: utf-8 -*-
# vispy: gallery 30
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
"""
Display an Isocurve
===================

Simple use of SceneCanvas to display an Isocurve visual.
"""
import sys
from vispy import app, scene, visuals
from vispy.util.filter import gaussian_filter
import numpy as np

canvas = scene.SceneCanvas(keys='interactive', title='Isocurve(s) overlayed '
                           'over Random Image Example')
canvas.size = 800, 600
canvas.show()

# Set up a viewbox to display the image with interactive pan/zoom
view = canvas.central_widget.add_view()

# Create the image
img_data = np.empty((200, 100, 3), dtype=np.ubyte)
noise = np.random.normal(size=(200, 100), loc=50, scale=150)
noise = gaussian_filter(noise, (4, 4, 0))
img_data[:] = noise[..., np.newaxis]
image = scene.visuals.Image(img_data, parent=view.scene)
# move image behind curves
image.transform = visuals.transforms.STTransform(translate=(0, 0, 0.5))

# level and color setup
levels = [40, 50, 60]
color_lev = [(1, 0, 0, 1),
             (1, 0.5, 0, 1),
             (1, 1, 0, 1)]

# Create isocurve, make a child of the image to ensure the two are always
# aligned.
curve = scene.visuals.Isocurve(noise, levels=levels, color_lev=color_lev,
                               parent=view.scene)

# Set 2D camera
view.camera = scene.PanZoomCamera(aspect=1)
# the camera will scale to the contents in the scene
view.camera.set_range()

if __name__ == '__main__' and sys.flags.interactive == 0:
    app.run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

"""
Scipt to generate code for our gl API, including all its backends.

The files involved in the code generation process are:

  * gl2.h - the C header file for the GL ES 2.0 API
  * vispy_ext.h - C header file with a subset of GL3+ constants defined
  * webgl.idl - the interface definition language for WebGL
  * annotations.py - manual annotations for non-trivial functions
  * headerparser.py - parses .h and .idl files 
  * createglapi.py - uses all of the above to generate the actual code

Rationale
---------

The GL ES 2.0 API is relatively small. Therefore a fully automated build
process like PyOpenGL now has is not really necessary. I tried to
automate what we can and simply do the rest with manual annotations.
Some groups of functions (like glUniform and friends) are handled in
*this* file.

Even though the API is quite small, we want to generate several
implementations, such as gl2 (desktop), es2 (angle on Windows), a generic
proxy, a mock backend and possibly more. Therefore automation
is crucial.

Further notes
-------------

This file is pretty big and even though I tried to make the code as clear
as possible, it's not always that easy to read. In effect this code is
not so easy to maintain. I hope it is at least clear enough so it can be
used to maintain the GL API itself.

This function must be run using Python3.
"""
from __future__ import annotations

import os
import sys

import headerparser
from annotations import parse_anotations

from OpenGL import GL  # For checking

THISDIR = os.path.abspath(os.path.dirname(__file__))
GLDIR = os.path.join(THISDIR, "..", "vispy", "gloo", "gl")

PREAMBLE = '''"""GL definitions converted to Python by codegen/createglapi.py.

THIS CODE IS AUTO-GENERATED. DO NOT EDIT.

%s

"""
'''


def create_parsers():
    # Create a parser for gl2 and es2
    parser1 = headerparser.Parser(os.path.join(THISDIR, "headers", "gl2.h"))
    parser2 = headerparser.Parser(os.path.join(THISDIR, "headers", "webgl.idl"))

    ## Check constants and generate API module

    # Get names
    names1 = set(parser1.constant_names)
    names2 = set(parser2.constant_names)

    # Check names correspondence
    if names1 == names2:
        print("Constants in gl2 and webgl are equal")
    else:
        print("===== Extra names in gl2 =====")
        print(", ".join(names1.difference(names2)))
        print("===== Extra names in webgl =====")
        print(", ".join(names2.difference(names1)))
        print("===========")

    # Test value correspondence
    superset = names1.intersection(names2)
    #
    constants = {}
    for c1 in parser1.constants.values():
        if c1.shortname in superset:
            constants[c1.shortname] = c1.value
    #
    assert len(constants) == len(superset)
    #
    for c2 in parser2.constants.values():
        if c2.shortname in constants:
            assert c2.value == constants[c2.shortname]
    print("Hooray! All constants that occur in both namespaces have equal values.")

    return parser1, parser2


DEFINE_ENUM = '''
class Enum(int):
    """Enum (integer) with a meaningfull repr."""
    
    def __new__(cls, name, value):
        base = int.__new__(cls, value)
        base.name = name
        return base
        
    def __repr__(self):
        return self.name
'''

DEFINE_CONST_MAP = '''
ENUM_MAP = {}
for var_name, ob in list(globals().items()):
    if var_name.startswith('GL_'):
        ENUM_MAP[int(ob)] = ob
del ob, var_name
'''


def create_constants_module(constant_definitions: list[headerparser.ConstantDefinition],
                            extension: bool = False) -> None:
    lines = [PREAMBLE % "Constants for OpenGL ES 2.0.", DEFINE_ENUM, ""]

    # For extensions, we only take the OES ones, and remove the OES
    if extension:
        renamed_defs = []
        for c in constant_definitions:
            if "OES" in c.oname:
                c.oname = c.oname.replace("OES", "")
                c.oname = c.oname.replace("__", "_").strip("_")
                renamed_defs.append(c)
        constant_definitions = renamed_defs

    # Insert constants
    for c in sorted(constant_definitions, key=lambda x: x.oname):
        if isinstance(c.value, int):
            lines.append("%s = Enum(%r, %r)" % (c.oname, c.oname, c.value))
        else:
            lines.append("%s = %r" % (c.oname, c.value))
    lines.append("")

    lines.append(DEFINE_CONST_MAP)

    # Write the file
    fname = "_constants_ext.py" if extension else "_constants.py"
    with open(os.path.join(GLDIR, fname), "wb") as f:
        f.write(("\n".join(lines)).encode("utf-8"))
    print("wrote %s" % fname)


IGNORE_FUNCTIONS = ["releaseShaderCompiler", "shaderBinary"]

WEBGL_EQUIVALENTS = {
    "genBuffers": "createBuffer",
    "genFramebuffers": "createFramebuffer",
    "genRenderbuffers": "createRenderbuffer",
    "genTextures": "createTexture",
    "deleteBuffers": "deleteBuffer",
    "deleteFramebuffers": "deleteFramebuffer",
    "deleteRenderbuffers": "deleteRenderbuffer",
    "deleteTextures": "deleteTexture",
    "clearDepthf": "clearDepth",
    "depthRangef": "depthRange",
    "getBufferParameteriv": "getBufferParameter",
    "getRenderbufferParameteriv": "getRenderbufferParameter",
    "getFramebufferAttachmentParameteriv": "getFramebufferAttachmentParameter",
    "getVertexAttribPointerv": "getVertexAttribOffset",
    "getProgramiv": "getProgramParameter",
    "getShaderiv": "getShaderParameter",
    "getBooleanv": "getParameter",
    "getFloatv": "getParameter",
    "getIntegerv": "getParameter",
    "getString": "getParameter",  # getParameter is getString + getFloat
}

# Types that we convert easily from Python to C (and back)
EASY_TYPES = {
    "void": (type(None), "c_voidp"),  # only for output
    "GLenum": (int, "c_uint"),
    "GLboolean": (bool, "c_bool"),
    "GLuint": (int, "c_uint"),
    "GLint": (int, "c_int"),
    "GLbitfield": (int, "c_uint"),
    "GLsizei": (int, "c_int"),
    "GLfloat": (float, "c_float"),
    "GLclampf": (float, "c_float"),
}

# Types that dont map 1-on-1 to Python values, but that we know
# how to set the ctypes argtypes for.
HARDER_TYPES = {
    "GLenum*": ("", "POINTER(ctypes.c_uint)"),
    "GLboolean*": ("", "POINTER(ctypes.c_bool)"),
    "GLuint*": ("", "POINTER(ctypes.c_uint)"),
    "GLint*": ("", "POINTER(ctypes.c_int)"),
    "GLsizei*": ("", "POINTER(ctypes.c_int)"),
    "GLfloat*": ("", "POINTER(ctypes.c_float)"),
    "GLubyte*": ("", "c_char_p"),
    "GLchar*": ("", "c_char_p"),
    "GLchar**": ("", "POINTER(ctypes.c_char_p)"),
    "GLvoid*": ("", "c_void_p"),  # or c_voidp?
    "GLvoid**": ("", "POINTER(ctypes.c_void_p)"),
    "GLintptr": ("", "c_ssize_t"),
    "GLsizeiptr": ("", "c_ssize_t"),
}

# Together the EASY_TYPES and HARDER_TYPES should cover all types that
# ES 2.0 uses.
KNOWN_TYPES = EASY_TYPES.copy()
KNOWN_TYPES.update(HARDER_TYPES)


def apiname(funcname):
    """Define what name the API uses, the short or the gl version."""
    if funcname.startswith("gl"):
        return funcname
    else:
        if funcname.startswith("_"):
            return "_gl" + funcname[1].upper() + funcname[2:]
        else:
            return "gl" + funcname[0].upper() + funcname[1:]


class FunctionDescription:
    def __init__(self, name, es2func, wglfunc, annfunc):
        self.name = name
        self.apiname = apiname(name)
        self.es2 = es2func
        self.wgl = wglfunc
        self.ann = annfunc
        self.args = []


class ApiGenerator:
    """Base API generator class. We derive several subclasses to implement
    the different backends.
    """

    backend_name = None

    DESCRIPTION = "GL API X"
    PREAMBLE = ""

    def __init__(self):
        self.lines = []
        self.functions_auto = set()
        self.functions_anno = set()
        self.functions_todo = set()

    def save(self):
        # Remove too many whitespace
        text = "\n".join(self.lines) + "\n"
        for i in range(10):
            text = text.replace("\n\n\n\n", "\n\n\n")
        # Write
        with open(self.filename, "wb") as f:
            f.write((PREAMBLE % self.DESCRIPTION).encode("utf-8"))
            for line in self.PREAMBLE.splitlines():
                f.write(line[4:].encode("utf-8") + b"\n")
            f.write(b"\n")
            f.write(text.encode("utf-8"))

    def add_functions(self, all_functions):
        for func_description in all_functions:
            self.add_function(func_description)

    def add_function(self, des: FunctionDescription):
        if des.es2.group:
            if des.name.startswith("get"):
                assert len(des.es2.group) == 2  # vi and fv
                des.es2 = des.es2.group[0]  # f comes before 1
                self._add_function(des)
            else:
                self._add_function_group(des)
        else:
            self._add_function(des)
        self.lines.append("\n")  # two lines between each function

    def _add_function_group(self, des: FunctionDescription) -> tuple[set, set, set]:
        lines = self.lines
        handled = True

        # Create map to es2 function objects
        es2funcs = {}
        for f in des.es2.group:
            cname = f.shortname
            es2funcs[cname] = f

        if des.name == "uniform":
            for t in ("float", "int"):
                for i in (1, 2, 3, 4):
                    args = ", ".join(["v%i" % j for j in range(1, i + 1)])
                    cname = "uniform%i%s" % (i, t[0])
                    sig = "%s(location, %s)" % (apiname(cname), args)
                    self._add_group_function(des, sig, es2funcs[cname])
            for t in ("float", "int"):
                for i in (1, 2, 3, 4):
                    cname = "uniform%i%sv" % (i, t[0])
                    sig = "%s(location, count, values)" % apiname(cname)
                    self._add_group_function(des, sig, es2funcs[cname])
        elif des.name == "uniformMatrix":
            for i in (2, 3, 4):
                cname = "uniformMatrix%ifv" % i
                sig = "%s(location, count, transpose, values)" % apiname(cname)
                self._add_group_function(des, sig, es2funcs[cname])
        elif des.name == "vertexAttrib":
            for i in (1, 2, 3, 4):
                args = ", ".join(["v%i" % j for j in range(1, i + 1)])
                cname = "vertexAttrib%if" % i
                sig = "%s(index, %s)" % (apiname(cname), args)
                self._add_group_function(des, sig, es2funcs[cname])
        elif des.name == "texParameter":
            for t in ("float", "int"):
                cname = "texParameter%s" % t[0]
                sig = "%s(target, pname, param)" % apiname(cname)
                self._add_group_function(des, sig, es2funcs[cname])
        else:
            handled = False

        if handled:
            self.functions_auto.add(des.name)
        else:
            self.functions_todo.add(des.name)
            lines.append("# todo: Dont know group %s" % des.name)

    def _add_function(self, des):
        # Need to be overloaded in subclass
        raise NotImplementedError()

    def _add_group_function(self, des, sig, es2func):
        # Need to be overloaded in subclass
        raise NotImplementedError()


class ProxyApiGenerator(ApiGenerator):
    """Generator for the general proxy class that will be loaded into gloo.gl."""

    filename = os.path.join(GLDIR, "_proxy.py")
    DESCRIPTION = "Base proxy API for GL ES 2.0."
    PREAMBLE = '''
    class BaseGLProxy(object):
        """ Base proxy class for the GL ES 2.0 API. Subclasses should
        implement __call__ to process the API calls.
        """
       
        def __call__(self, funcname, returns, *args):
            raise NotImplementedError()
    '''

    def _returns(self, des):
        shortame = des.name
        for prefix in ("get", "is", "check", "create", "read"):
            if shortame.startswith(prefix):
                return True
        else:
            return False

    def _add_function(self, des):
        ret = self._returns(des)
        prefix = "return " if ret else ""
        argstr = ", ".join(des.args)
        self.lines.append("    def %s(self, %s):" % (des.apiname, argstr))
        self.lines.append(
            '        %sself("%s", %r, %s)' % (prefix, apiname(des.name), ret, argstr)
        )

    def _add_group_function(self, des, sig, es2func):
        ret = self._returns(des)
        prefix = "return " if ret else ""
        funcname = apiname(sig.split("(")[0])
        args = sig.split("(", 1)[1].split(")")[0]
        self.lines.append("    def %s(self, %s):" % (funcname, args))
        self.lines.append(
            '        %sself("%s", %r, %s)' % (prefix, funcname, ret, args)
        )


class Gl2ApiGenerator(ApiGenerator):
    """Generator for the gl2 (desktop) backend."""

    filename = os.path.join(GLDIR, "_gl2.py")
    write_c_sig = True
    define_argtypes_in_module = False
    backend_name = "gl"

    DESCRIPTION = "Subset of desktop GL API compatible with GL ES 2.0"
    PREAMBLE = """
    import ctypes
    from .gl2 import _lib, _get_gl_func
    """

    def _get_argtype_str(self, es2func):
        ct_arg_types = [KNOWN_TYPES.get(arg.ctype, None) for arg in es2func.args]
        # Set argument types on ctypes function
        if None in ct_arg_types:
            argstr = "UNKNOWN_ARGTYPES"
        elif es2func.group:
            argstr = "UNKNOWN_ARGTYPES"
        else:
            argstr = ", ".join(["ctypes.%s" % t[1] for t in ct_arg_types[1:]])
            argstr = "()" if not argstr else "(%s,)" % argstr
        # Set output arg (if available)
        if ct_arg_types[0][0] != type(None):
            resstr = "ctypes.%s" % ct_arg_types[0][1]
        else:
            resstr = "None"
        return resstr, argstr

    def _write_argtypes(self, es2func):
        lines = self.lines
        ct_arg_types = [KNOWN_TYPES.get(arg.ctype, None) for arg in es2func.args]
        # Set argument types on ctypes function
        if None in ct_arg_types:
            lines.append("# todo: unknown argtypes")
        elif es2func.group:
            lines.append("# todo: oops, dont set argtypes for group!")
        else:
            if ct_arg_types[1:]:
                argstr = ", ".join(["ctypes.%s" % t[1] for t in ct_arg_types[1:]])
                lines.append("_lib.%s.argtypes = %s," % (es2func.glname, argstr))
            else:
                lines.append("_lib.%s.argtypes = ()" % es2func.glname)
        # Set output arg (if available)
        if ct_arg_types[0][0] != type(None):
            lines.append(
                "_lib.%s.restype = ctypes.%s" % (es2func.glname, ct_arg_types[0][1])
            )

    def _native_call_line(self, name, es2func, cargstr=None, prefix="", indent=4):
        resstr, argstr = self._get_argtype_str(es2func)
        if cargstr is None:
            cargs = [arg.name for arg in es2func.args[1:]]
            cargstr = ", ".join(cargs)

        lines = "try:\n"
        lines += "    nativefunc = %s._native\n" % apiname(name)
        lines += "except AttributeError:\n"
        lines += '    nativefunc = %s._native = _get_gl_func("%s", %s, %s)\n' % (
            apiname(name),
            es2func.glname,
            resstr,
            argstr,
        )
        lines += "%snativefunc(%s)\n" % (prefix, cargstr)
        lines = [" " * indent + line for line in lines.splitlines()]
        return "\n".join(lines)

    def _add_function(self, des):
        lines = self.lines
        es2func = des.es2

        # Write arg types
        if self.define_argtypes_in_module:
            self._write_argtypes(es2func)

        # Get names and types of C-API
        ce_arg_types = [arg.ctype for arg in es2func.args[1:]]
        ce_arg_names = [arg.name for arg in es2func.args[1:]]
        ct_arg_types = [KNOWN_TYPES.get(arg.ctype, None) for arg in es2func.args]
        ct_arg_types_easy = [EASY_TYPES.get(arg.ctype, None) for arg in es2func.args]

        # Write C function signature, for debugging and development
        if self.write_c_sig:
            argnamesstr = ", ".join(
                [
                    c_type + " " + c_name
                    for c_type, c_name in zip(ce_arg_types, ce_arg_names)
                ]
            )
            lines.append(
                "# %s = %s(%s)" % (es2func.args[0].ctype, es2func.oname, argnamesstr)
            )

        # Write Python function def
        lines.append("def %s(%s):" % (des.apiname, ", ".join(des.args)))

        # Construct C function call
        cargs = [arg.name for arg in des.es2.args[1:]]

        # Now write the body of the function ...
        if des.ann:
            prefix = "res = "
            # Annotation available
            self.functions_anno.add(des.name)
            callline = self._native_call_line(des.name, es2func, prefix=prefix)
            lines.extend(des.ann.get_lines(callline, self.backend_name))

        elif es2func.group:
            # Group?
            self.functions_todo.add(des.name)
            lines.append("    pass  # todo: Oops. this is a group!")
        elif None in ct_arg_types_easy:
            self.functions_todo.add(des.name)
            lines.append("    pass  # todo: Not all easy types!")
        elif des.args != [arg.name for arg in des.wgl.args[1:]]:
            self.functions_todo.add(des.name)
            lines.append("    pass  # todo: ES 2.0 and WebGL args do not match!")
        else:
            # This one is easy!
            self.functions_auto.add(des.name)
            # Get prefix
            prefix = ""
            if ct_arg_types[0][0] != type(None):
                prefix = "return "
            elif des.es2.shortname.startswith("get"):
                raise RuntimeError("Get func returns void?")
            # Set string
            callline = self._native_call_line(des.name, des.es2, prefix=prefix)
            lines.append(callline)

        if "gl2" in self.__class__.__name__.lower():
            # Post-fix special cases for gl2. See discussion in #201
            # glDepthRangef and glClearDepthf are not always available,
            # and sometimes they do not work if they are
            if es2func.oname in ("glDepthRangef", "glClearDepthf"):
                for i in range(1, 10):
                    line = lines[-i]
                    if not line.strip() or line.startswith("#"):
                        break
                    line = line.replace("c_float", "c_double")
                    line = line.replace("glDepthRangef", "glDepthRange")
                    line = line.replace("glClearDepthf", "glClearDepth")
                    lines[-i] = line

    def _add_group_function(self, des, sig, es2func):
        lines = self.lines

        call_line = self._native_call_line

        if self.define_argtypes_in_module:
            self._write_argtypes(es2func)

        funcname = sig.split("(", 1)[0]
        args = sig.split("(", 1)[1].split(")")[0]

        if des.name == "uniform":
            if funcname[-1] != "v":
                lines.append("def %s:" % sig)
                lines.append(call_line(funcname, es2func, args))
            else:
                t = {"f": "float", "i": "int"}[funcname[-2]]
                lines.append("def %s:" % sig)
                lines.append("    values = [%s(val) for val in values]" % t)
                lines.append("    values = (ctypes.c_%s*len(values))(*values)" % t)
                lines.append(call_line(funcname, es2func, "location, count, values"))
        elif des.name == "uniformMatrix":
            lines.append("def %s:" % sig)
            lines.append('    if not values.flags["C_CONTIGUOUS"]:')
            lines.append("        values = values.copy()")
            lines.append('    assert values.dtype.name == "float32"')
            lines.append("    values_ = values")
            lines.append(
                "    values = values_.ctypes.data_as(ctypes.POINTER(ctypes.c_float))"
            )
            lines.append(
                call_line(funcname, es2func, "location, count, transpose, values")
            )
        elif des.name == "vertexAttrib":
            lines.append("def %s:" % sig)
            lines.append(call_line(funcname, es2func, args))
        elif des.name == "texParameter":
            lines.append("def %s:" % sig)
            lines.append(call_line(funcname, es2func, args))

        else:
            raise ValueError("unknown group func")


class Es2ApiGenerator(Gl2ApiGenerator):
    """Generator for the es2 backend (i.e. Angle on Windows). Very
    similar to the gl2 API, but we do not need that deferred loading
    of GL functions here.
    """

    filename = os.path.join(GLDIR, "_es2.py")
    write_c_sig = True
    define_argtypes_in_module = True
    backend_name = "es"

    DESCRIPTION = "GL ES 2.0 API (via Angle/DirectX on Windows)"
    PREAMBLE = """
    import ctypes
    from .es2 import _lib
    """

    def _native_call_line(self, name, es2func, cargstr=None, prefix="", indent=4):
        resstr, argstr = self._get_argtype_str(es2func)
        if cargstr is None:
            cargs = [arg.name for arg in es2func.args[1:]]
            cargstr = ", ".join(cargs)
        return " " * indent + "%s_lib.%s(%s)" % (prefix, es2func.glname, cargstr)


class PyOpenGL2ApiGenerator(ApiGenerator):
    """Generator for a fallback pyopengl backend."""

    filename = os.path.join(GLDIR, "_pyopengl2.py")
    backend_name = "pyopengl"

    DESCRIPTION = "Proxy API for GL ES 2.0 subset, via the PyOpenGL library."
    PREAMBLE = """
    import ctypes
    from OpenGL import GL
    import OpenGL.GL.framebufferobjects as FBO
    """

    def __init__(self):
        ApiGenerator.__init__(self)
        self._functions_to_import = []
        self._used_functions = []

    def _add_function(self, des):
        # Fix for FBO?
        mod = "GL"
        if "renderbuffer" in des.name.lower() or "framebuffer" in des.name.lower():
            mod = "FBO"
        # Get call line
        argstr = ", ".join(des.args)
        call_line = "    return %s.%s(%s)" % (mod, des.es2.glname, argstr)
        # Get annotation lines
        ann_lines = []
        if des.ann is not None:
            ann_lines = des.ann.get_lines(call_line, self.backend_name)
        # Use annotation or not
        if ann_lines:
            self.lines.append("def %s(%s):" % (des.apiname, argstr))
            self.lines.extend(ann_lines)
            self._used_functions.append(des.es2.glname)
        else:
            # To be imported from OpenGL.GL
            self._functions_to_import.append((des.es2.glname, des.apiname))

    def _add_group_function(self, des, sig, es2func):
        # All group functions can be directly imported from OpenGL
        funcname = apiname(sig.split("(")[0])
        self._functions_to_import.append((funcname, funcname))

    def save(self):
        # Write remaining functions
        self.lines.append("# List of functions that we should import from OpenGL.GL")
        self.lines.append("_functions_to_import = [")
        for name1, name2 in self._functions_to_import:
            self.lines.append('    ("%s", "%s"),' % (name1, name2))
        self.lines.append("    ]")

        self.lines.append("")

        # Write used functions
        self.lines.append("# List of functions in OpenGL.GL that we use")
        self.lines.append("_used_functions = [")
        for name in self._used_functions:
            self.lines.append('    "%s",' % name)
        self.lines.append("    ]")

        # Really save
        ApiGenerator.save(self)


class FunctionCollector:
    def __init__(self, annotations, parser1, parser2):
        self.annotations = annotations
        self.parser1 = parser1
        self.parser2 = parser2
        # Keep track of what webGL names we "used"
        self.used_webgl_names = set()

        # Also keep track of what functions we could handle automatically,
        # and which not. Just for reporting.
        self.functions_auto = set()
        self.functions_anno = set()
        self.functions_todo = set()

        self.all_functions = []

    def collect_function_definitions(self):
        """Process function definitions of ES 2.0, WebGL and annotations.

        We try to combine information from these three sources to find the
        arguments for the Python API. In this "merging" process we also
        check for inconsistencies between the API definitions.

        """
        for name in self.parser1.function_names:
            if name in IGNORE_FUNCTIONS:
                continue

            # Get es2 function
            es2func = self.parser1.functions[name]

            # Get webgl version
            lookupname = WEBGL_EQUIVALENTS.get(es2func.shortname, es2func.shortname)
            wglfunc = self.parser2.functions.get(lookupname, None)
            if wglfunc:
                self.used_webgl_names.add(lookupname)
            else:
                print("WARNING: %s not available in WebGL" % es2func.shortname)

            # Convert name
            name = WEBGL_EQUIVALENTS.get(name, name)

            # Avoid duplicates for getParameter
            if name == "getParameter":
                if es2func.shortname != "getString":
                    name = "_" + es2func.shortname

            # Get annotated version
            annfunc = self.annotations.get(name, None)

            # Create description instance
            des = FunctionDescription(name, es2func, wglfunc, annfunc)
            self.all_functions.append(des)

            # Get information about arguments
            argnames_es2 = [arg.name for arg in es2func.args[1:]]
            if wglfunc:
                argnames_wgl = [arg.name for arg in wglfunc.args[1:]]
            if annfunc:
                argnames_ann = annfunc.args  # Can contain 'argname=default'
                argnames_ann = [arg.split("=")[0] for arg in argnames_ann]

            # Set argumenets specification of our GL API
            # Also check and report when we deviate from the WebGL API
            if wglfunc and argnames_es2 == argnames_wgl:
                if annfunc and argnames_ann != argnames_es2:
                    des.args = argnames_ann
                    print(
                        "WARNING: %s: Annotation overload even though webgl and es2 match."
                        % name
                    )
                else:
                    des.args = argnames_es2
            elif wglfunc:
                if annfunc and argnames_ann != argnames_wgl:
                    des.args = argnames_ann
                    print("WARNING: %s: Annotation overload webgl args." % name)
                else:
                    # print('WARNING: %s: assuming wgl args.'%name)
                    des.args = argnames_wgl
            else:
                print("WARNING: %s: Could not determine args!!" % name)

            # Go over all functions to test if they are in OpenGL
            for func in [es2func, wglfunc]:
                if func is None:
                    continue
                group = func.group or [func]
                for f in group:
                    # Check opengl
                    if f.oname.startswith("gl") and not hasattr(GL, f.glname):
                        print("WARNING: %s seems not available in PyOpenGL" % f.glname)

    def check_unused_webgl_funcs(self):
        """Check which WebGL functions we did not find/use."""
        for name in set(self.parser2.function_names).difference(self.used_webgl_names):
            print("WARNING: WebGL function %s not in Desktop" % name)

    def report_status(self):
        print(
            "Could generate %i functions automatically, and %i with annotations"
            % (len(self.functions_auto), len(self.functions_anno))
        )
        print("Need more info for %i functions." % len(self.functions_todo))
        if not self.functions_todo:
            print("Hooray! All %i functions are covered!" % len(self.all_functions))


def main():
    annotations = parse_anotations()
    gl2es2_parser, webgl_parser = create_parsers()
    vispy_ext_parser = headerparser.Parser(os.path.join(THISDIR, "headers", "vispy_ext.h"))
    constant_definitions = list(gl2es2_parser.constants.values()) + list(vispy_ext_parser.constants.values())
    create_constants_module(constant_definitions)

    # Get full function definitions and report
    func_collector = FunctionCollector(annotations, gl2es2_parser, webgl_parser)
    func_collector.collect_function_definitions()

    for Gen in [
        ProxyApiGenerator,
        Gl2ApiGenerator,
        Es2ApiGenerator,
        PyOpenGL2ApiGenerator,
    ]:
        gen = Gen()
        gen.add_functions(func_collector.all_functions)
        func_collector.functions_auto.update(gen.functions_auto)
        func_collector.functions_anno.update(gen.functions_anno)
        func_collector.functions_todo.update(gen.functions_todo)
        gen.save()

    func_collector.check_unused_webgl_funcs()
    func_collector.report_status()


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
# vispy: gallery 2

"""
Markers with Instanced Rendering
================================

Compare instanced and GL_POINTS rendering methods for Markers visual.

This example shows how instanced rendering works around platform point size limits
(many platforms limit GL_POINTS to 64px or smaller) and demonstrates the
canvas_size_limits feature for constraining marker sizes during zoom/pan.

Controls:
* m: Toggle between 'points' and 'instanced' rendering methods
* s: Cycle through marker sizes
* c: Toggle canvas size clamping (min 10px, max 100px)
* z: Toggle scaling mode: 'fixed' vs 'scene' (grows/shrinks with zoom)
* l: Toggle spherical lighting (3D sphere effect)
* k: Cycle through marker shapes (disc, arrow, ring, etc.)

Notes:

* you may not see a difference between methods on your platform - that's fine!
* 'instanced' method should be the same across platforms (arbitrarily large markers)
* Canvas size clamping keeps markers readable during zoom
* Spherical lighting adds depth to markers with simulated 3D lighting
* There may be lighting direction differences between methods (known issue)
"""

from itertools import cycle

from vispy import scene, use
from vispy.visuals.markers import symbol_shaders

use(gl="gl+")


marker_sizes = cycle([16, 32, 64, 96, 128, 256, 512])
scaling_modes = cycle(["fixed", "scene"])
marker_symbols = cycle(symbol_shaders.keys())


class Canvas(scene.SceneCanvas):
    def __init__(self):
        scene.SceneCanvas.__init__(
            self, keys="interactive", size=(512, 512), title="Instanced Markers Demo"
        )
        self.unfreeze()
        self.view = self.central_widget.add_view()
        self.view.camera = scene.PanZoomCamera(rect=(0, 0, 512, 512), aspect=1.0)

        self.marker_positions, self.face_colors = _create_markers_pattern()
        self.method = "instanced"
        self.current_size = next(marker_sizes)
        self.current_symbol = next(marker_symbols)
        self.clamping_enabled = False
        self.scaling_mode = next(scaling_modes)
        self.spherical_enabled = False
        self.markers = scene.visuals.Markers(
            method=self.method,
            parent=self.view.scene,
            scaling=self.scaling_mode,
            spherical=self.spherical_enabled,
        )

        self.freeze()

        self.markers.set_data(
            self.marker_positions,
            face_color=self.face_colors,
            edge_color="black",
            size=self.current_size,
            edge_width=2,
            symbol=self.current_symbol,
        )

        self.view.bgcolor = "#2e3440"

        self.print_state()
        self.show()

    def print_state(self, changed=None):
        """Print current state with optional highlighting of what changed."""
        clamp_str = (
            f"{self.markers.canvas_size_limits}"
            if self.clamping_enabled
            else "off"
        )

        parts = {
            'method': f"method={self.method}",
            'symbol': f"symbol={self.current_symbol}",
            'size': f"size={self.current_size}px",
            'clamp': f"clamp={clamp_str}",
            'scaling': f"scaling={self.scaling_mode}",
            'lighting': f"lighting={'on' if self.spherical_enabled else 'off'}",
        }

        # highlight the changed part in bold
        if changed and changed in parts:
            parts[changed] = f"\033[1m{parts[changed]}\033[0m"

        state_line = " | ".join(parts.values())
        state_line = state_line.ljust(120)
        print(f"\r{state_line}", end="", flush=True)

    def on_key_press(self, event):
        if event.text == "m":
            self.method = "instanced" if self.method == "points" else "points"

            # recreate markers with new method, cannot change method on the fly
            self.markers.parent = None
            self.markers = scene.visuals.Markers(
                method=self.method, parent=self.view.scene
            )
            self.markers.set_data(
                self.marker_positions,
                face_color=self.face_colors,
                edge_color="black",
                size=self.current_size,
                edge_width=2,
                symbol=self.current_symbol,
            )
            self.markers.scaling = self.scaling_mode
            self.markers.spherical = self.spherical_enabled
            if self.clamping_enabled:
                self.markers.canvas_size_limits = (10, 100)

            self.print_state(changed='method')
            self.update()

        elif event.text == "s":
            self.current_size = next(marker_sizes)
            self.markers.set_data(
                self.marker_positions,
                face_color=self.face_colors,
                edge_color="black",
                size=self.current_size,
                edge_width=2,
                symbol=self.current_symbol,
            )
            self.print_state(changed='size')
            self.update()

        elif event.text == "c":
            self.clamping_enabled = not self.clamping_enabled
            if self.clamping_enabled:
                self.markers.canvas_size_limits = (10, 100)
            else:
                self.markers.canvas_size_limits = None
            self.print_state(changed='clamp')
            self.update()

        elif event.text == "z":
            self.scaling_mode = next(scaling_modes)
            self.markers.scaling = self.scaling_mode
            self.print_state(changed='scaling')
            self.update()

        elif event.text == "l":
            self.spherical_enabled = not self.spherical_enabled
            self.markers.spherical = self.spherical_enabled
            self.print_state(changed='lighting')
            self.update()

        elif event.text == "k":
            self.current_symbol = next(marker_symbols)
            self.markers.symbol = self.current_symbol
            self.print_state(changed='symbol')
            self.update()


def _create_markers_pattern():
    import numpy as np

    # Create positions in a circle plus one in the center
    n = 12
    angles = np.linspace(0, 2 * np.pi, n, endpoint=False)
    radius = 150
    center = np.array([256, 256])
    pos = np.column_stack(
        [center[0] + radius * np.cos(angles), center[1] + radius * np.sin(angles)]
    ).astype(np.float32)

    pos = np.vstack([pos, center])

    colors = np.zeros((n + 1, 4), dtype=np.float32)
    for i in range(n):
        hue = i / n
        h = hue * 6
        x = 1 - abs(h % 2 - 1)
        if h < 1:
            colors[i] = [1, x, 0, 1]
        elif h < 2:
            colors[i] = [x, 1, 0, 1]
        elif h < 3:
            colors[i] = [0, 1, x, 1]
        elif h < 4:
            colors[i] = [0, x, 1, 1]
        elif h < 5:
            colors[i] = [x, 0, 1, 1]
        else:
            colors[i] = [1, 0, x, 1]
    colors[-1] = [1, 1, 1, 1]

    return pos, colors


if __name__ == "__main__":
    from vispy import app
    print(__doc__)
    canvas = Canvas()
    app.run()



# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

""" Code to parse a header file and create a list of constants,
functions (with arguments). This information can then be used to
autogenerate our OpenGL API.
"""

import os


def getwords(line):
    """Get words on a line."""
    line = line.replace("\t", " ").strip()
    return [w for w in line.split(" ") if w]


class Parser:
    """Class to parse header files.

    It can deal with gl2.h and webgl.idl,
    as well as some desktop OpenGL header files. It produces a list of
    ConstantDefinition objects and FunctionDefinition objects, which can
    be accessed via a dict.

    """

    def __init__(self, header_file, parse_now=True):
        # Get filenames for C and Py
        self._c_fname = os.path.split(header_file)[1]

        # Get absolute filenames
        self._c_filename = header_file

        # Init intermediate results
        self._functionDefs = []
        self._constantDefs = []

        # Init output
        self._functions = {}
        self._constants = {}
        # cache of constant values for constant aliases
        self._constant_values = {}

        # We are aware of the line number
        self._linenr = 0

        # Some stats
        self.stat_types = set()

        if parse_now:
            self.parse()

    def parse(self):
        """Parse the header file!"""
        self._parse_constants_and_functions_from_file()

        # Remove invalid defs
        self._functionDefs = [d for d in self._functionDefs if d.isvalid]
        self._constantDefs = [d for d in self._constantDefs if d.isvalid]

        # Collect multipe similar functions in groups
        self._functionDefs.sort(key=lambda x: x.glname)
        keyDef = None
        keyDefs = []
        for funcDef in [f for f in self._functionDefs]:
            # Check if we need a new keydef
            if funcDef.extrachars:
                # Create new keydef or use old one?
                if keyDef and keyDef.glname == funcDef.keyname:
                    pass  # Keep same keydef
                else:
                    keyDef = FunctionGroup.from_function_def(funcDef)
                    keyDefs.append(keyDef)
                # Add to group
                keyDef.group.append(funcDef)
        # Process function groups
        for keyDef in keyDefs:
            if len(keyDef.group) > 1:
                self._functionDefs.append(keyDef)
                for d in keyDef.group:
                    self._functionDefs.remove(d)

        # Sort constants and functions
        self._functionDefs.sort(key=lambda x: x.glname)
        self._constantDefs.sort(key=lambda x: x.glname)

        # Get dicts
        for definition in self._functionDefs:
            self._functions[definition.shortname] = definition
        for definition in self._constantDefs:
            self._constants[definition.shortname] = definition

        # Get some stats
        for funcDef in self._functionDefs:
            for arg in funcDef.args:
                self.stat_types.add(arg.ctype)

        # Show stats
        # TODO: Remove iteration
        n1 = len([d for d in self._constantDefs])
        n2 = len([d for d in self._functionDefs])
        n3 = len([d for d in self._functionDefs if d.group])
        n4 = sum([len(d.group) for d in self._functionDefs if d.group])
        print(
            'Found %i constants and %i unique functions (%i groups contain %i functions)").'
            % (n1, n2, n3, n4)
        )

        print("C-types found in args:", self.stat_types)

    def _parse_constants_and_functions_from_file(self):
        line_gen = self._get_nonblank_lines()
        for line in line_gen:
            if line.startswith(("#define", "const GLenum")):
                const_def = ConstantDefinition(line)
                const_def.parse_line(self._constant_values)
                self._append_definition(const_def)
            elif "(" in line:
                while ")" not in line:
                    line += next(line_gen)
                if line.endswith(");"):
                    func_def = FunctionDefinition(line)
                    func_def.parse_line()
                    self._append_definition(func_def)

    def _get_nonblank_lines(self):
        with open(self._c_filename, "rt", encoding="utf-8") as header_file:
            for line in header_file:
                line = line.strip()
                if line:
                    yield line
                self._linenr += 1

    def _append_definition(self, definition):
        """Add an output line. Can be multiple lines."""
        # Create comment
        definition.comment = "line %i of %s" % (self._linenr, self._c_fname)
        # Add to lists
        if isinstance(definition, FunctionDefinition):
            self._functionDefs.append(definition)
        elif isinstance(definition, ConstantDefinition):
            self._constantDefs.append(definition)

    @property
    def constant_names(self):
        """Sorted list of constant names."""
        return [d.shortname for d in self._constantDefs]

    @property
    def function_names(self):
        """Sorted list of function names."""
        return [d.shortname for d in self._functionDefs]

    @property
    def constants(self):
        """Dict with all the constants."""
        return self._constants

    @property
    def functions(self):
        """Dict witj all the functions."""
        return self._functions

    def show_groups(self):
        for d in self._functionDefs:
            if isinstance(d.group, list):
                print(d.keyname)
                for d2 in d.group:
                    print("  ", d2.glname)


class Definition:
    """Abstract class to represent a constant or function definition."""

    def __init__(self, line):
        self.line = line
        self.isvalid = True
        self.comment = ""
        self.oname = ""  # original name
        self.shortname = self.glname = ""  # short and long name

    def parse_line(self):
        # Do initial parsing of the incoming line (which may be multiline, actually)
        raise NotImplementedError()

    def _set_name(self, name):
        # Store original name
        self.oname = name
        # Store plain name
        if name.startswith("GL_"):
            name = name[3:]
        elif name.startswith("gl"):
            name = name[2].lower() + name[3:]
        self.shortname = name
        # Store gl name
        if name.upper() == name:
            name = "GL_" + name
        else:
            name = "gl" + name[0].upper() + name[1:]
        self.glname = name


class ConstantDefinition(Definition):
    def __init__(self, line):
        self.value = None
        super().__init__(line)

    def parse_line(self, existing_constants):
        """Set cname and value attributes."""
        value = None
        line = self.line.split("/*", 1)[0]
        args = getwords(line)[1:]
        self.isvalid = False
        if len(args) == 1:
            pass
        elif len(args) == 2:
            # Set name
            name, val = args
            self.isvalid = bool(name)
            self._set_name(name)
            value = self._set_value_from_string(val, existing_constants)
        elif "=" in args:
            name, val = args[-3], args[-1]
            self.isvalid = bool(name)
            self._set_name(name)
            value = self._set_value_from_string(val, existing_constants)
        else:
            print('Dont know what to do with "%s"' % line)

        # For when this constant is reused to set another constant
        if value is not None:
            existing_constants[self.oname] = value
        self.value = value

    def _set_value_from_string(self, val, existing_constants):
        # Set value
        val = val.strip(";")
        if val.startswith("0x"):
            value = int(val[2:].rstrip("ul"), 16)
        elif val[0] in "0123456789":
            value = int(val)
        elif val.startswith("'"):
            value = val
        elif val in existing_constants:
            value = existing_constants[val]
        else:
            print('Warning: Dont know what to do with "%s"' % val)
            value = None
        return value


class FunctionDefinition(Definition):

    SKIPTYPECHARS = "if"  # 'bsifd'
    ALLSKIPCHARS = SKIPTYPECHARS + "v1234"

    def __init__(self, line):
        self.keyname = None
        self.extrachars = None
        self.group = None
        self.cargs = []
        self.args = []
        super().__init__(line)

    def parse_line(self):
        """Set cname, keyname, cargs attributes.
        The list of args always has one entry and the first entry is always
        the output (can be void).
        """
        # Parse components
        beforeBrace, args = self.line.split("(", 1)
        betweenBraces, _ = args.split(")", 1)
        outs = getwords(beforeBrace)
        prefix, name = outs[:-1], outs[-1]

        # Store name
        self._set_name(name)

        # Possibly, this function belongs to a collection of similar functions,
        # which we are going to replace with one function in Python.
        self.keyname = self.glname.rstrip("v").rstrip(self.SKIPTYPECHARS).rstrip("1234")
        self.extrachars = self.matchKeyName(self.keyname)

        # If this is a list, this instance represents the group
        # If this is True, this instance is in a group (but not the
        # representative)
        self.group = None

        # Create list of Argument instances
        self.cargs = [arg.strip() for arg in betweenBraces.split(",")]
        self.args = []
        # Set output arg
        self.args.append(Argument(" ".join(prefix), False))
        # Parse input arguments,
        for arg in self.cargs:
            if arg and arg != "void":
                self.args.append(Argument(arg))

    def matchKeyName(self, keyname):
        if self.glname.startswith(keyname):
            extrachars = self.glname[len(keyname) :]
            if all([(c in self.ALLSKIPCHARS) for c in extrachars]):
                return extrachars


class FunctionGroup(FunctionDefinition):
    @classmethod
    def from_function_def(cls, func_def):
        keyDef = cls(func_def.line)
        keyDef.parse_line()
        keyDef._set_name(func_def.keyname)
        return keyDef

    def parse_line(self):
        super().parse_line()
        self.group = []  # must be set after line is parsed


class Argument:
    def __init__(self, argAsString, cinput=True):
        # Parse string
        components = [c for c in argAsString.split(" ") if c]
        if len(components) == 1:
            name = "unknown_name"
            type = components[0]
        else:
            name = components[-1]
            type = components[-2]
            if "const" in type:
                type = components[
                    -3
                ]  # glShaderSource has "const GLchar* const* string"
        # Store stuff
        self.orig = tuple(components)
        self.name = name.lstrip("*")
        self.isptr = argAsString.count("*")  # Number of stars
        self.ctype = type.strip("*") + "*" * self.isptr
        # Status flags
        self.cinput = cinput


if __name__ == "__main__":
    THISDIR = os.path.abspath(os.path.dirname(__file__))

    # Some tests ...
    gl2 = Parser(os.path.join(THISDIR, "headers", "gl2.h"))
    import OpenGL.GL

    pygl = set([name for name in dir(OpenGL.GL)])

    # Test if all functions are in pyopengl
    for keyfunc in gl2._functionDefs:
        group = keyfunc.group or [keyfunc]
        for f in group:
            if f.glname not in pygl:
                print("Not in pyopengl:", f.glname)

    # Test if constant match with these in pyopengl
    for d in gl2._constantDefs:
        v1 = d.value
        try:
            v2 = getattr(OpenGL.GL, d.glname)
        except AttributeError:
            print(d.glname, "is not in pyopengl")
        else:
            if v1 != v2:
                print(d.glname, "does not match: %r vs %r" % (v1, int(v2)))



"""
Demonstration of Tube
"""

import sys
from vispy import scene
from vispy.geometry.torusknot import TorusKnot

from colorsys import hsv_to_rgb
import numpy as np

canvas = scene.SceneCanvas(keys='interactive')
view = canvas.central_widget.add_view()

points1 = TorusKnot(5, 3).first_component[:-1]
points1[:, 0] -= 20.
points1[:, 2] -= 15.

points2 = points1.copy()
points2[:, 2] += 30.

points3 = points1.copy()
points3[:, 0] += 41.
points3[:, 2] += 30

points4 = points1.copy()
points4[:, 0] += 41.

points5 = points1.copy()
points5[:, 0] += 20.4
points5[:, 2] += 15

colors = np.linspace(0, 1, len(points1))
colors = np.array([hsv_to_rgb(c, 1, 1) for c in colors])

vertex_colors = np.random.random(8 * len(points1))
vertex_colors = np.array([hsv_to_rgb(c, 1, 1) for c in vertex_colors])

l1 = scene.visuals.Tube(points1,
                        shading='flat',
                        color=colors,  # this is overridden by
                                       # the vertex_colors argument
                        vertex_colors=vertex_colors,
                        tube_points=8)

l2 = scene.visuals.Tube(points2,
                        color=['red', 'green', 'blue'],
                        shading='smooth',
                        tube_points=8)

l3 = scene.visuals.Tube(points3,
                        color=colors,
                        shading='flat',
                        tube_points=8,
                        closed=True)

l4 = scene.visuals.Tube(points4,
                        color=colors,
                        shading='smooth',
                        tube_points=8,
                        mode='lines')

# generate sine wave radii
radii = np.sin(2 * np.pi * 440 * np.arange(points5.shape[0]) / 44000)
radii = (radii + 1.5) / 2

l5 = scene.visuals.Tube(points5,
                        radius=radii,
                        color='white',
                        shading='smooth',
                        closed=True,
                        tube_points=8)

view.add(l1)
view.add(l2)
view.add(l3)
view.add(l4)
view.add(l5)
view.camera = scene.TurntableCamera()
# tube does not expose its limits yet
view.camera.set_range((-20, 20), (-20, 20), (-20, 20))
canvas.show()

if __name__ == '__main__':
    if sys.flags.interactive != 1:
        canvas.app.run()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
"""
Tutorial: Creating Visuals
==========================

03. Antialiasing
----------------

In [tutorial 1] we learned how to draw a simple rectangle, and in [tutorial 2]
we expanded on this by using the Document coordinate system to draw a 
rectangular border of a specific width. In this tutorial we introduce the
Framebuffer coordinate system, which is used for antialiasing measurements. 

In order to antialias our edges, we need to introduce a calculation to the
fragment shader that computes, for each pixel being drawn, the fraction of the 
pixel that is covered by the visual's geometry. At first glance, it may seem
that the Document coordinate system is sufficient for this purpose because it
has unit-length pixels. However, there are two situations when the actual 
pixels being filled by the fragment shader are not the same size as the pixels
on the canvas:

    1. High-resolution displays (such as retina displays) that report a canvas
       resolution smaller than the actual framebuffer resolution.
    2. When exporting to an image with a different size than the canvas.

In most cases the discrepancy between Document and Framebuffer coordinates can
be corrected by a simple scale factor. However, this fails for some interesting
corner cases where the transform is more complex, such as in VR applications 
using optical distortion correction. Decide for yourself: is this Visual for 
my personal use, or is it intended for a broader audience? For simplicity in 
this example, we will use a simple scale factor.
"""

from vispy import app, gloo, visuals, scene
import numpy as np


# Here we use almost the same vertex shader as in tutorial 2.
# The important difference is the addition of the line_pos variable,
# which measures position across the width of the border line.
vertex_shader = """
varying float line_pos;  // how far we are across the border line

void main() {
    // First map the vertex to document coordinates
    vec4 doc_pos = $visual_to_doc(vec4($position, 0, 1));
    
    vec4 adjusted;
    if ( $adjust_dir.x == 0. ) {
        adjusted = doc_pos;
        line_pos = $line_width;  // at the outside of the border
    }
    else {
        // Inner vertexes must be adjusted for line width
        vec4 doc_x = $visual_to_doc(vec4($adjust_dir.x, 0, 0, 0)) - 
                    $visual_to_doc(vec4(0, 0, 0, 0));
        vec4 doc_y = $visual_to_doc(vec4(0, $adjust_dir.y, 0, 0)) - 
                    $visual_to_doc(vec4(0, 0, 0, 0));
        doc_x = normalize(doc_x);
        doc_y = normalize(doc_y);
                        
        vec4 proj_y_x = dot(doc_x, doc_y) * doc_x;  // project y onto x
        float cur_width = length(doc_y - proj_y_x);  // measure current weight
        
        // And now we can adjust vertex position for line width:
        adjusted = doc_pos + ($line_width / cur_width) * (doc_x + doc_y);
        
        line_pos = 0;  // at the inside of the border
    }
    
    // Finally map the remainder of the way to render coordinates
    gl_Position = $doc_to_render(adjusted);
}
"""

# The fragment shader is updated to change the opacity of the color based on
# the amount of the fragment that is covered by the visual's geometry.
fragment_shader = """
varying float line_pos;

void main() {
    // Decrease the alpha linearly as we come within 1 pixel of the edge.
    // Note: this only approximates the actual fraction of the pixel that is
    // covered by the visual's geometry. A more accurate measurement would
    // produce better antialiasing, but the effect would be subtle.
    float alpha = 1.0;
    if ((line_pos * $doc_fb_scale) < 1) {
        alpha = $color.a * line_pos;
    }
    else if ((line_pos * $doc_fb_scale) > ($line_width - 1)) {
        alpha = $color.a * ($line_width - line_pos);
    }
    gl_FragColor = vec4($color.rgb, alpha);
}
"""


# The visual class is defined almost exactly as in [tutorial 2]. The only 
# major difference is that the draw() method now calculates a scale factor
# for converting between document and framebuffer coordinates.
class MyRectVisual(visuals.Visual):
    """Visual that draws a rectangular outline.

    Parameters
    ----------
    x : float
        x coordinate of rectangle origin
    y : float
        y coordinate of rectangle origin
    w : float
        width of rectangle
    h : float
        height of rectangle
    weight : float
        width of border (in px)
    """

    def __init__(self, x, y, w, h, weight=4.0):
        self.weight = weight
        visuals.Visual.__init__(self, vertex_shader, fragment_shader)

        # 10 vertices for 8 triangles (using triangle_strip) forming a 
        # rectangular outline
        self.vert_buffer = gloo.VertexBuffer(np.array([
            [x, y], 
            [x, y], 
            [x+w, y], 
            [x+w, y], 
            [x+w, y+h],
            [x+w, y+h],
            [x, y+h],
            [x, y+h],
            [x, y], 
            [x, y], 
        ], dtype=np.float32))

        # Direction each vertex should move to correct for line width
        # (the length of this vector will be corrected in the shader)
        self.adj_buffer = gloo.VertexBuffer(np.array([
            [0, 0],
            [1, 1],
            [0, 0],
            [-1, 1],
            [0, 0],
            [-1, -1],
            [0, 0],
            [1, -1],
            [0, 0],
            [1, 1],
        ], dtype=np.float32))

        self.shared_program.vert['position'] = self.vert_buffer
        self.shared_program.vert['adjust_dir'] = self.adj_buffer
        # To compensate for antialiasing, add 1 to border width:
        self.shared_program.vert['line_width'] = weight + 1
        self.shared_program.frag['color'] = (1, 0, 0, 1)
        self._draw_mode = 'triangle_strip'
        self.set_gl_state(cull_face=False)

    def _prepare_transforms(self, view):
        # Set the two transforms required by the vertex shader:
        tr = view.transforms
        view_vert = view.view_program.vert
        view_vert['visual_to_doc'] = tr.get_transform('visual', 'document')
        view_vert['doc_to_render'] = tr.get_transform('document', 'render')

        # Set the scale factor between document and framebuffer coordinate
        # systems. This assumes a simple linear / isotropic scale; more complex
        # transforms will yield strange results!
        doc_to_fb = tr.get_transform('document', 'framebuffer')
        fbs = np.linalg.norm(doc_to_fb.map([1, 0]) - doc_to_fb.map([0, 0]))
        view_frag = view.view_program.frag
        view_frag['doc_fb_scale'] = fbs
        view_frag['line_width'] = (self.weight + 1) * fbs

# As in the previous tutorial, we auto-generate a Visual+Node class for use
# in the scenegraph.
MyRect = scene.visuals.create_visual_node(MyRectVisual)


# Finally we will test the visual by displaying in a scene.

canvas = scene.SceneCanvas(keys='interactive', show=True)

# This time we add a ViewBox to let the user zoom/pan
view = canvas.central_widget.add_view()
view.camera = 'panzoom'
view.camera.rect = (0, 0, 800, 800)

# ..and add the rects to the view instead of canvas.scene
rects = [MyRect(100, 100, 200, 300, parent=view.scene),
         MyRect(500, 100, 200, 300, parent=view.scene)]

# Again, rotate one rectangle to ensure the transforms are working as we 
# expect.
tr = visuals.transforms.MatrixTransform()
tr.rotate(25, (0, 0, 1))
rects[1].transform = tr

# Add some text instructions
text = scene.visuals.Text("Drag right mouse button to zoom.", 
                          color='w',
                          anchor_x='left',
                          parent=view,
                          pos=(20, 30))

# ..and optionally start the event loop
if __name__ == '__main__':
    import sys
    if sys.flags.interactive != 1:
        app.run()



# -*- coding: utf-8 -*-
# vispy: testskip
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
"""
This is a simple egl example that produces 2d plot.

Was tested on Linux with Nvidia drivers.

Requires libEGLso.1 to be found in the system.
"""
import vispy
import numpy as np
import vispy.plot as vp
import vispy.io as io
vispy.use(app='egl')

# Check the application correctly picked up egl 
assert vispy.app.use_app().backend_name == 'egl', 'Not using EGL'

if __name__ == '__main__':

    data = np.load(io.load_data_file('electrophys/iv_curve.npz'))['arr_0']
    time = np.arange(0, data.shape[1], 1e-4)

    fig = vp.Fig(size=(800, 800), show=False)

    x = np.linspace(0, 10, 20)
    y = np.cos(x)
    line = fig[0, 0].plot((x, y), symbol='o', width=3, title='I/V Curve',
                          xlabel='Current (pA)', ylabel='Membrane Potential (mV)')
    grid = vp.visuals.GridLines(color=(0, 0, 0, 0.5))
    grid.set_gl_state('translucent')
    fig[0, 0].view.add(grid)

    fig.show()

    img = fig.render()
    io.write_png("egl.png", img)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vispy: testskip

"""
Example demonstrating the use of two GLCanvases in one QtApp.
"""

from PyQt5 import QtCore, QtWidgets
import sys

from fireworks import Canvas as FireCanvas
from rain import Canvas as RainCanvas


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        QtWidgets.QMainWindow.__init__(self)

        self.resize(1000, 500)
        self.setWindowTitle('vispy example ...')

        self.splitter_h = QtWidgets.QSplitter(QtCore.Qt.Horizontal)

        # Central Widget
        splitter1 = QtWidgets.QSplitter(QtCore.Qt.Horizontal)

        self.rain_canvas = RainCanvas()
        self.rain_canvas.create_native()
        self.rain_canvas.native.setParent(self)
        self.fireworks_canvas = FireCanvas()
        self.fireworks_canvas.create_native()
        self.fireworks_canvas.native.setParent(self)

        splitter1.addWidget(self.fireworks_canvas.native)
        splitter1.addWidget(self.rain_canvas.native)

        self.setCentralWidget(splitter1)


if __name__ == '__main__':
    appQt = QtWidgets.QApplication(sys.argv)
    win = MainWindow()
    win.show()
    appQt.exec_()



# -*- coding: utf-8 -*-
# vispy: gallery 30
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
"""
Demonstration of borders and background colors
==============================================
"""

from vispy.scene import SceneCanvas

canvas = SceneCanvas(keys='interactive', bgcolor='w', show=True)
grid = canvas.central_widget.add_grid(spacing=0, bgcolor='gray',
                                      border_color='k')
view1 = grid.add_view(row=0, col=0, margin=10, bgcolor=(1, 0, 0, 0.5),
                      border_color=(1, 0, 0))
view2 = grid.add_view(row=0, col=1, margin=10, bgcolor=(0, 1, 0, 0.5),
                      border_color=(0, 1, 0))

if __name__ == '__main__':
    canvas.app.run()



# -*- coding: utf-8 -*-
# vispy: gallery 2
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
"""
More complex grid layout
========================

::

    +---+-------+---+
    | 1 |   1   | 2 |
    |---+-------+---+
    | 3 | Empty | 2 |
    +---+-------+---+
    | 3 |   4   | 4 |
    +---+-------+---+

"""

import sys

from vispy import scene, app

canvas = scene.SceneCanvas(keys='interactive')
canvas.size = 600, 600
canvas.bgcolor = "#000000"
canvas.show()

grid = canvas.central_widget.add_grid()

# top_left
grid.add_widget(row=0, col=0, col_span=2, bgcolor="#ffffff")

# top_right
grid.add_widget(row=0, col=2, row_span=2, bgcolor="#dddddd")

# bottom_left
grid.add_widget(row=1, col=0, row_span=2, bgcolor="#444444")


# bottom_right
grid.add_widget(row=2, col=1, col_span=2, bgcolor="#888888")


if __name__ == '__main__' and sys.flags.interactive == 0:
    app.run()



# -*- coding: utf-8 -*-
# vispy: gallery 10
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

"""
Create a Point Cloud
====================

Demonstrates use of visual.Markers to create a point cloud with a
standard turntable camera to fly around with and a centered 3D Axis.
"""

import numpy as np
import vispy.scene
from vispy.scene import visuals

#
# Make a canvas and add simple view
#
canvas = vispy.scene.SceneCanvas(keys='interactive', show=True)
view = canvas.central_widget.add_view()


# generate data
pos = np.random.normal(size=(100000, 3), scale=0.2)
# one could stop here for the data generation, the rest is just to make the
# data look more interesting. Copied over from magnify.py
centers = np.random.normal(size=(50, 3))
indexes = np.random.normal(size=100000, loc=centers.shape[0] / 2,
                           scale=centers.shape[0] / 3)
indexes = np.clip(indexes, 0, centers.shape[0] - 1).astype(int)
symbols = np.random.choice(['o', '^'], len(pos))
scales = 10**(np.linspace(-2, 0.5, centers.shape[0]))[indexes][:, np.newaxis]
pos *= scales
pos += centers[indexes]

# create scatter object and fill in the data
scatter = visuals.Markers()
scatter.set_data(pos, edge_width=0, face_color=(1, 1, 1, .5), size=5, symbol=symbols)

view.add(scatter)

view.camera = 'turntable'  # or try 'arcball'

# add a colored 3D axis for orientation
axis = visuals.XYZAxis(parent=view.scene)

if __name__ == '__main__':
    import sys
    if sys.flags.interactive != 1:
        vispy.app.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vispy: gallery 2
"""
Draw 2D points
==============

Simple example plotting 2D points.

"""

from vispy import gloo
from vispy import app
import numpy as np

VERT_SHADER = """
attribute vec2  a_position;
attribute vec3  a_color;
attribute float a_size;

varying vec4 v_fg_color;
varying vec4 v_bg_color;
varying float v_radius;
varying float v_linewidth;
varying float v_antialias;

void main (void) {
    v_radius = a_size;
    v_linewidth = 1.0;
    v_antialias = 1.0;
    v_fg_color  = vec4(0.0,0.0,0.0,0.5);
    v_bg_color  = vec4(a_color,    1.0);

    gl_Position = vec4(a_position, 0.0, 1.0);
    gl_PointSize = 2.0*(v_radius + v_linewidth + 1.5*v_antialias);
}
"""

FRAG_SHADER = """
#version 120

varying vec4 v_fg_color;
varying vec4 v_bg_color;
varying float v_radius;
varying float v_linewidth;
varying float v_antialias;
void main()
{
    float size = 2.0*(v_radius + v_linewidth + 1.5*v_antialias);
    float t = v_linewidth/2.0-v_antialias;
    float r = length((gl_PointCoord.xy - vec2(0.5,0.5))*size);
    float d = abs(r - v_radius) - t;
    if( d < 0.0 )
        gl_FragColor = v_fg_color;
    else
    {
        float alpha = d/v_antialias;
        alpha = exp(-alpha*alpha);
        if (r > v_radius)
            gl_FragColor = vec4(v_fg_color.rgb, alpha*v_fg_color.a);
        else
            gl_FragColor = mix(v_bg_color, v_fg_color, alpha);
    }
}
"""


class Canvas(app.Canvas):

    def __init__(self):
        app.Canvas.__init__(self, keys='interactive')
        ps = self.pixel_scale

        # Create vertices
        n = 10000
        v_position = 0.25 * np.random.randn(n, 2).astype(np.float32)
        v_color = np.random.uniform(0, 1, (n, 3)).astype(np.float32)
        v_size = np.random.uniform(2*ps, 12*ps, (n, 1)).astype(np.float32)

        self.program = gloo.Program(VERT_SHADER, FRAG_SHADER)
        # Set uniform and attribute
        self.program['a_color'] = gloo.VertexBuffer(v_color)
        self.program['a_position'] = gloo.VertexBuffer(v_position)
        self.program['a_size'] = gloo.VertexBuffer(v_size)
        gloo.set_state(clear_color='white', blend=True,
                       blend_func=('src_alpha', 'one_minus_src_alpha'))

        self.show()

    def on_resize(self, event):
        gloo.set_viewport(0, 0, *event.physical_size)

    def on_draw(self, event):
        gloo.clear(color=True, depth=True)
        self.program.draw('points')


if __name__ == '__main__':
    canvas = Canvas()
    app.run()



# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

""" 
This module contains manual annotations for the gl backends. Together
with the header files, we can generatre the full ES 2.0 API.

Every function-annotations consists of sections that apply to one or
more backends. If no backends are specified in the first section, it
applies to all backends.
"""

import ctypes


# bind / gen / delete stuff


def deleteBuffer(buffer):
    # --- gl es
    n = 1  
    buffers = (ctypes.c_uint*n)(buffer)  
    ()  
    # --- pyopengl
    GL.glDeleteBuffers(1, [buffer])


def deleteFramebuffer(framebuffer):
    # --- gl es
    n = 1  
    framebuffers = (ctypes.c_uint*n)(framebuffer)  
    ()
    # --- pyopengl
    FBO.glDeleteFramebuffers(1, [framebuffer])


def deleteRenderbuffer(renderbuffer):
    # --- gl es
    n = 1  
    renderbuffers = (ctypes.c_uint*n)(renderbuffer)  
    ()
    # --- pyopengl
    FBO.glDeleteRenderbuffers(1, [renderbuffer])


def deleteTexture(texture):
    # --- gl es
    n = 1  
    textures = (ctypes.c_uint*n)(texture)  
    ()
    # --- pyopengl
    GL.glDeleteTextures([texture])


def createBuffer():
    # --- gl es
    n = 1
    buffers = (ctypes.c_uint*n)()
    ()  
    return buffers[0]
    # --- pyopengl
    return GL.glGenBuffers(1)
    # --- mock
    return 1


def createFramebuffer():
    # --- gl es
    n = 1
    framebuffers = (ctypes.c_uint*n)()
    ()
    return framebuffers[0]
    # --- pyopengl
    return FBO.glGenFramebuffers(1)
    # --- mock
    return 1


def createRenderbuffer():
    # --- gl es
    n = 1
    renderbuffers = (ctypes.c_uint*n)()
    ()
    return renderbuffers[0]
    # --- pyopengl
    return FBO.glGenRenderbuffers(1)
    # --- mock
    return 1


def createTexture():
    # --- gl es
    n = 1
    textures = (ctypes.c_uint*n)()
    ()
    return textures[0]
    # --- pyopengl
    return GL.glGenTextures(1)
    # --- mock
    return 1


# Image stuff
def texImage2D(target, level, internalformat, format, type, pixels):
    border = 0
    # --- gl es
    if isinstance(pixels, (tuple, list)):
        height, width = pixels
        pixels = ctypes.c_void_p(0)
        pixels = None
    else:
        if not pixels.flags['C_CONTIGUOUS']:
            pixels = pixels.copy('C')
        pixels_ = pixels
        pixels = pixels_.ctypes.data
        height, width = pixels_.shape[:2]
    ()
    # --- pyopengl
    if isinstance(pixels, (tuple, list)):
        height, width = pixels
        pixels = None
    else:
        height, width = pixels.shape[:2]
    GL.glTexImage2D(target, level, internalformat, width, height, border, format, type, pixels)
    

def texSubImage2D(target, level, xoffset, yoffset, format, type, pixels):
    # --- gl es
    if not pixels.flags['C_CONTIGUOUS']:
        pixels = pixels.copy('C')
    pixels_ = pixels
    pixels = pixels_.ctypes.data
    height, width = pixels_.shape[:2]
    ()
    # --- pyopengl
    height, width = pixels.shape[:2]
    GL.glTexSubImage2D(target, level, xoffset, yoffset, width, height, format, type, pixels)


def readPixels(x, y, width, height, format, type):
    # --- es
    # GL_ALPHA, GL_RGB, GL_RGBA
    t = {6406:1, 6407:3, 6408:4}[format]
    # --- gl mock
    # GL_ALPHA, GL_RGB, GL_RGBA, GL_DEPTH_COMPONENT
    t = {6406:1, 6407:3, 6408:4, 6402:1}[format]
    # --- gl es mock
    # GL_UNSIGNED_BYTE, GL_FLOAT
    nb = {5121:1, 5126:4}[type]
    size = int(width*height*t*nb)
    # --- gl es
    pixels = ctypes.create_string_buffer(size)
    ()
    return pixels[:]
    # --- mock
    return size * b'\x00'


def compressedTexImage2D(target, level, internalformat, width, height, border=0, data=None):
    # border = 0  # set in args
    # --- gl es
    if not data.flags['C_CONTIGUOUS']:
        data = data.copy('C')
    data_ = data
    size = data_.size
    data = data_.ctypes.data
    ()
    # --- pyopengl
    size = data.size
    GL.glCompressedTexImage2D(target, level, internalformat, width, height, border, size, data)


def compressedTexSubImage2D(target, level, xoffset, yoffset, width, height, format, data):
    # --- gl es
    if not data.flags['C_CONTIGUOUS']:
        data = data.copy('C')
    data_ = data
    size = data_.size
    data = data_.ctypes.data
    ()
    # --- pyopengl
    size = data.size
    GL.glCompressedTexSubImage2D(target, level, xoffset, yoffset, width, height, format, size, data)


# Buffer data


def bufferData(target, data, usage):
    """Data can be numpy array or the size of data to allocate."""
    # --- gl es
    if isinstance(data, int):
        size = data
        data = ctypes.c_voidp(0)
    else:
        if not data.flags['C_CONTIGUOUS'] or not data.flags['ALIGNED']:
            data = data.copy('C')
        data_ = data
        size = data_.nbytes
        data = data_.ctypes.data
    ()
    # --- pyopengl
    if isinstance(data, int):
        size = data
        data = None
    else:
        size = data.nbytes
    GL.glBufferData(target, size, data, usage)


def bufferSubData(target, offset, data):
    # --- gl es
    if not data.flags['C_CONTIGUOUS']:
        data = data.copy('C')
    data_ = data
    size = data_.nbytes
    data = data_.ctypes.data
    ()
    # --- pyopengl
    size = data.nbytes
    GL.glBufferSubData(target, offset, size, data)


def drawElements(mode, count, type, offset):
    # --- gl es
    if offset is None:
        offset = ctypes.c_void_p(0)
    elif isinstance(offset, ctypes.c_void_p):
        pass
    elif isinstance(offset, (int, ctypes.c_int)):
        offset = ctypes.c_void_p(int(offset))
    else:
        if not offset.flags['C_CONTIGUOUS']:
            offset = offset.copy('C')
        offset_ = offset
        offset = offset.ctypes.data
    indices = offset
    ()
    # --- pyopengl
    if offset is None:
        offset = ctypes.c_void_p(0)
    elif isinstance(offset, (int, ctypes.c_int)):
        offset = ctypes.c_void_p(int(offset))
    ()


def vertexAttribPointer(indx, size, type, normalized, stride, offset):
    # --- gl es
    if offset is None:
        offset = ctypes.c_void_p(0)
    elif isinstance(offset, ctypes.c_void_p):
        pass
    elif isinstance(offset, (int, ctypes.c_int)):
        offset = ctypes.c_void_p(int(offset))
    else:
        if not offset.flags['C_CONTIGUOUS']:
            offset = offset.copy('C')
        offset_ = offset
        offset = offset.ctypes.data
        # We need to ensure that the data exists at draw time :(
        # PyOpenGL does this too
        key = '_vert_attr_'+str(indx)
        setattr(glVertexAttribPointer, key, offset_)
    ptr = offset
    ()
    # --- pyopengl
    if offset is None:
        offset = ctypes.c_void_p(0)
    elif isinstance(offset, (int, ctypes.c_int)):
        offset = ctypes.c_void_p(int(offset))
    ()


def bindAttribLocation(program, index, name):
    # --- gl es
    name = ctypes.c_char_p(name.encode('utf-8'))
    ()
    # --- pyopengl
    name = name.encode('utf-8')
    ()


# Setters


def shaderSource(shader, source):
    # Some implementation do not like getting a list of single chars
    if isinstance(source, (tuple, list)):
        strings = [s for s in source]
    else:
        strings = [source]
    # --- gl es
    count = len(strings)  
    string = (ctypes.c_char_p*count)(*[s.encode('utf-8') for s in strings])  
    length = (ctypes.c_int*count)(*[len(s) for s in strings])  
    ()
    # --- pyopengl
    GL.glShaderSource(shader, strings)


# Getters

def _getBooleanv(pname):
    # --- gl es
    params = (ctypes.c_bool*1)()
    ()
    return params[0]


def _getIntegerv(pname):
    # --- gl es
    n = 16
    d = -2**31  # smallest 32bit integer
    params = (ctypes.c_int*n)(*[d for i in range(n)])
    ()
    params = [p for p in params if p!=d]
    if len(params) == 1:
        return params[0]
    else:
        return tuple(params)


def _getFloatv(pname):
    # --- gl es
    n = 16
    d = float('Inf')
    params = (ctypes.c_float*n)(*[d for i in range(n)])
    ()
    params = [p for p in params if p!=d]
    if len(params) == 1:
        return params[0]
    else:
        return tuple(params)

# def _getString(pname):
#     # --- gl es
#     ()
#     return res.value
#     # --- mock
#     return ''


def getParameter(pname):
    if pname in [33902, 33901, 32773, 3106, 2931, 2928, 
                 2849, 32824, 10752, 32938]:
        # GL_ALIASED_LINE_WIDTH_RANGE GL_ALIASED_POINT_SIZE_RANGE
        # GL_BLEND_COLOR GL_COLOR_CLEAR_VALUE GL_DEPTH_CLEAR_VALUE
        # GL_DEPTH_RANGE GL_LINE_WIDTH GL_POLYGON_OFFSET_FACTOR
        # GL_POLYGON_OFFSET_UNITS GL_SAMPLE_COVERAGE_VALUE
        # --- gl es
        return _glGetFloatv(pname)
        # --- pyopengl
        return GL.glGetFloatv(pname)
        # /---
    elif pname in [7936, 7937, 7938, 35724, 7939]:
        # GL_VENDOR, GL_RENDERER, GL_VERSION, GL_SHADING_LANGUAGE_VERSION, 
        # GL_EXTENSIONS are strings
        pass  # string handled below
    else:
        # --- gl es
        return _glGetIntegerv(pname)
        # --- pyopengl
        return GL.glGetIntegerv(pname)
        # /---
    # --- gl es
    name = pname
    ()
    return ctypes.string_at(res).decode('utf-8') if res else ''
    # --- pyopengl
    res = GL.glGetString(pname)
    return res.decode('utf-8')


def getUniform(program, location):
    # --- gl es
    n = 16
    d = float('Inf')
    params = (ctypes.c_float*n)(*[d for i in range(n)])
    ()
    params = [p for p in params if p!=d]
    if len(params) == 1:
        return params[0]
    else:
        return tuple(params)
    # --- pyopengl
    n = 16
    d = float('Inf')
    params = (ctypes.c_float*n)(*[d for i in range(n)])
    GL.glGetUniformfv(program, location, params)
    params = [p for p in params if p!=d]
    if len(params) == 1:
        return params[0]
    else:
        return tuple(params)


def getVertexAttrib(index, pname):
    # --- gl es
    n = 4
    d = float('Inf')
    params = (ctypes.c_float*n)(*[d for i in range(n)])
    ()
    params = [p for p in params if p!=d]
    if len(params) == 1:
        return params[0]
    else:
        return tuple(params)
    # --- pyopengl
    # From PyOpenGL v3.1.0 the glGetVertexAttribfv(index, pname) does
    # work, but it always returns 4 values, with zeros in the empty
    # spaces. We have no way to tell whether they are empty or genuine
    # zeros. Fortunately, pyopengl also supports the old syntax.
    n = 4
    d = float('Inf')
    params = (ctypes.c_float*n)(*[d for i in range(n)])
    GL.glGetVertexAttribfv(index, pname, params)
    params = [p for p in params if p!=d]
    if len(params) == 1:
        return params[0]
    else:
        return tuple(params)


def getTexParameter(target, pname):
    # --- gl es
    d = float('Inf')
    params = (ctypes.c_float*1)(d)
    ()
    return params[0]


def getActiveAttrib(program, index):
    # --- gl es pyopengl
    bufsize = 256
    # --- gl es
    length = (ctypes.c_int*1)()
    size = (ctypes.c_int*1)()
    type = (ctypes.c_uint*1)()
    name = ctypes.create_string_buffer(bufsize)
    # --- gl es
    ()
    name = name[:length[0]].decode('utf-8')
    return name, size[0], type[0]
    # --- pyopengl
    name, size, type = GL.glGetActiveAttrib(program, index, bufSize=bufsize)
    return name.decode('utf-8'), size, type
    # --- mock
    return 'mock_val', 1, 5126


def getVertexAttribOffset(index, pname):
    # --- gl es
    pointer = (ctypes.c_void_p*1)()
    ()
    return pointer[0] or 0
    # --- pyopengl
    try:  # maybe the fixed it
        ()
    except TypeError:
        pointer = (ctypes.c_void_p*1)()
        GL.glGetVertexAttribPointerv(index, pname, pointer)
        return pointer[0] or 0
    # --- mock
    return 0

    
def getActiveUniform(program, index):
    # --- gl es
    bufsize = 256
    length = (ctypes.c_int*1)()
    size = (ctypes.c_int*1)()
    type = (ctypes.c_uint*1)()
    name = ctypes.create_string_buffer(bufsize)
    ()
    name = name[:length[0]].decode('utf-8')
    return name, size[0], type[0]
    # --- pyopengl
    name, size, type = GL.glGetActiveUniform(program, index)
    return name.decode('utf-8'), size, type


def getAttachedShaders(program):
    # --- gl es
    maxcount = 256
    count = (ctypes.c_int*1)()
    shaders = (ctypes.c_uint*maxcount)()
    ()
    return tuple(shaders[:count[0]])


def getAttribLocation(program, name):
    # --- gl es
    name = ctypes.c_char_p(name.encode('utf-8'))
    ()
    return res
    # --- pyopengl
    name = name.encode('utf-8')
    ()
    

def getUniformLocation(program, name):
    # --- gl es
    name = ctypes.c_char_p(name.encode('utf-8'))
    ()
    return res
    # --- pyopengl
    name = name.encode('utf-8')
    ()


def getProgramInfoLog(program):
    # --- gl es
    bufsize = 1024
    length = (ctypes.c_int*1)()
    infolog = ctypes.create_string_buffer(bufsize)
    ()
    return infolog[:length[0]].decode('utf-8')
    # --- pyopengl
    res = GL.glGetProgramInfoLog(program)
    return res.decode('utf-8') if isinstance(res, bytes) else res


def getShaderInfoLog(shader):
    # --- gl es
    bufsize = 1024
    length = (ctypes.c_int*1)()
    infolog = ctypes.create_string_buffer(bufsize)
    ()
    return infolog[:length[0]].decode('utf-8')
    # --- pyopengl
    res = GL.glGetShaderInfoLog(shader)
    return res.decode('utf-8') if isinstance(res, bytes) else res


def getProgramParameter(program, pname):
    # --- gl es
    params = (ctypes.c_int*1)()
    ()
    return params[0]


def getShaderParameter(shader, pname):
    # --- gl es
    params = (ctypes.c_int*1)()
    ()
    return params[0]


def getShaderPrecisionFormat(shadertype, precisiontype):
    # --- gl es
    range = (ctypes.c_int*1)()
    precision = (ctypes.c_int*1)()
    ()
    return range[0], precision[0]


def getShaderSource(shader):
    # --- gl es
    bufsize = 1024*1024
    length = (ctypes.c_int*1)()
    source = (ctypes.c_char*bufsize)()
    ()
    return source.value[:length[0]].decode('utf-8')
    # --- pyopengl
    res = GL.glGetShaderSource(shader)
    return res.decode('utf-8')
    

def getBufferParameter(target, pname):
    # --- gl es
    d = -2**31  # smallest 32bit integer
    params = (ctypes.c_int*1)(d)
    ()
    return params[0]


def getFramebufferAttachmentParameter(target, attachment, pname):
    # --- gl es
    d = -2**31  # smallest 32bit integer
    params = (ctypes.c_int*1)(d)
    ()
    return params[0]
    # --- pyopengl
    d = -2**31  # smallest 32bit integer
    params = (ctypes.c_int*1)(d)
    FBO.glGetFramebufferAttachmentParameteriv(target, attachment, pname, params)
    return params[0]


def getRenderbufferParameter(target, pname):
    # --- gl es
    d = -2**31  # smallest 32bit integer
    params = (ctypes.c_int*1)(d)
    ()
    return params[0]
    # --- pyopengl
    d = -2**31  # smallest 32bit integer
    params = (ctypes.c_int*1)(d)
    FBO.glGetRenderbufferParameteriv(target, pname, params)
    return params[0]



## ============================================================================


class FunctionAnnotation:
    def __init__(self, name, args, output):
        self.name = name
        self.args = args
        self.output = output
        self.lines = []  # (line, comment) tuples
    
    def __repr__(self):
        return '<FunctionAnnotation for %s>' % self.name
        
    def get_lines(self, call, backend):
        """ Get the lines for this function based on the given backend. 
        The given API call is inserted at the correct location.
        """
        if backend is None:
            raise RuntimeError("Backend must be specified!")
        backend_selector = (backend, )  # first lines are for all backends
        lines = []
        for line in self.lines:
            if line.lstrip().startswith('# ---'):
                backend_selector = line.strip().split(' ')
                continue
            if line.lstrip().startswith('# /---'):
                backend_selector = backend
                continue
            if backend in backend_selector:
                if line.strip() == '()':
                    indent = line.split('(')[0][4:]
                    line = indent + call
                lines.append(line)
        return lines
    
    def is_arg_set(self, name):
        """Get whether a given variable name is set.

        This allows checking whether a variable that is an input to the C
        function is not an input for the Python function, and may be an output.

        """
        needle = '%s =' % name
        for line, comment in self.lines:
            if line.startswith(needle):
                return True
        else:
            return False


def parse_anotations():
    """Parse this annotations file and produce a dictionary of FunctionAnnotation objects."""
    functions = {}
    function = None
    
    for line in open(__file__, 'rt').readlines():
        # Stop?
        if '='*40 in line:
            break
        
        if line.startswith('def '):
            name = line.split(' ')[1].split('(')[0]
            args = line.split('(')[1].split(')')[0].split(', ')
            args = [arg for arg in args if arg]
            out = line.partition('->')[2].strip()
            function = FunctionAnnotation(name, args, out)
            functions[name] = function
            continue
        elif not function:
            continue
        
        # Add line
        line = line.rstrip()
        indent = len(line) - len(line.strip())
        if line.strip() and indent >= 4:
            function.lines.append(line)

    return functions


if __name__ == '__main__':
    print(parse_anotations().keys())
    



# -*- coding: utf-8 -*-
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

"""Utilities for Vispy. A collection of modules that are used in
one or more Vispy sub-packages.
"""

from .logs import logger, set_log_level, use_log_level  # noqa
from .config import (config, sys_info, save_config, get_config_keys,  # noqa
                     set_data_dir, _TempDir, _get_args)  # noqa
from .fetching import load_data_file  # noqa
from .frozen import Frozen  # noqa
from . import fonts       # noqa
from . import transforms  # noqa
from .wrappers import use, run_subprocess  # noqa
from .bunch import SimpleBunch  # noqa

from typing import Optional

import numpy as np

# `copy` keyword semantics changed in NumPy 2.0
# this maintains compatibility with older versions
# if/when NumPy 2.0 becomes the minimum version, we can remove this
# we don't worry about early dev versions of NumPy 2.0 (that may or may not have the kwarg) here
# see:
#  https://numpy.org/devdocs/numpy_2_0_migration_guide.html#adapting-to-changes-in-the-copy-keyword
#  https://github.com/scipy/scipy/pull/20172
np_copy_if_needed: Optional[bool] = None
if np.lib.NumpyVersion(np.__version__) < "1.28.0":
    np_copy_if_needed = False



# -*- coding: utf-8 -*-
# vispy: gallery 30
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
"""
Shading a Mesh
==============

Show mesh filter usage for shading (lighting) a mesh and displaying a wireframe.
"""

import argparse

from vispy import app, scene
from vispy.io import read_mesh, load_data_file
from vispy.scene.visuals import Mesh
from vispy.scene import transforms
from vispy.visuals.filters import ShadingFilter, WireframeFilter


parser = argparse.ArgumentParser()
default_mesh = load_data_file('orig/triceratops.obj.gz')
parser.add_argument('--mesh', default=default_mesh)
parser.add_argument('--shininess', default=100)
parser.add_argument('--wireframe-width', default=1)
args, _ = parser.parse_known_args()

vertices, faces, normals, texcoords = read_mesh(args.mesh)

canvas = scene.SceneCanvas(keys='interactive', bgcolor='white')
view = canvas.central_widget.add_view()

view.camera = 'arcball'
view.camera.depth_value = 1e3

# Create a colored `MeshVisual`.
mesh = Mesh(vertices, faces, color=(.5, .7, .5, 1))
mesh.transform = transforms.MatrixTransform()
mesh.transform.rotate(90, (1, 0, 0))
mesh.transform.rotate(-45, (0, 0, 1))
view.add(mesh)

# Use filters to affect the rendering of the mesh.
wireframe_filter = WireframeFilter(width=args.wireframe_width)
# Note: For convenience, this `ShadingFilter` would be created automatically by
# the `MeshVisual with, e.g. `mesh = MeshVisual(..., shading='smooth')`. It is
# created manually here for demonstration purposes.
shading_filter = ShadingFilter(shininess=args.shininess)
# The wireframe filter is attached before the shading filter otherwise the
# wireframe is not shaded.
mesh.attach(wireframe_filter)
mesh.attach(shading_filter)


def attach_headlight(view):
    light_dir = (0, 1, 0, 0)
    shading_filter.light_dir = light_dir[:3]
    initial_light_dir = view.camera.transform.imap(light_dir)

    @view.scene.transform.changed.connect
    def on_transform_change(event):
        transform = view.camera.transform
        shading_filter.light_dir = transform.map(initial_light_dir)[:3]


attach_headlight(view)

shading_states = (
    dict(shading=None),
    dict(shading='flat'),
    dict(shading='smooth'),
)
shading_state_index = shading_states.index(
    dict(shading=shading_filter.shading))

wireframe_states = (
    dict(wireframe_only=False, faces_only=False,),
    dict(wireframe_only=True, faces_only=False,),
    dict(wireframe_only=False, faces_only=True,),
)
wireframe_state_index = wireframe_states.index(dict(
    wireframe_only=wireframe_filter.wireframe_only,
    faces_only=wireframe_filter.faces_only,
))


def cycle_state(states, index):
    new_index = (index + 1) % len(states)
    return states[new_index], new_index


@canvas.events.key_press.connect
def on_key_press(event):
    global shading_state_index
    global wireframe_state_index
    if event.key == 's':
        state, shading_state_index = cycle_state(shading_states,
                                                 shading_state_index)
        for attr, value in state.items():
            setattr(shading_filter, attr, value)
        mesh.update()
    elif event.key == 'w':
        wireframe_filter.enabled = not wireframe_filter.enabled
        mesh.update()
    elif event.key == 'f':
        state, wireframe_state_index = cycle_state(wireframe_states,
                                                   wireframe_state_index)
        for attr, value in state.items():
            setattr(wireframe_filter, attr, value)
        mesh.update()


canvas.show()


if __name__ == "__main__":
    app.run()



# -*- coding: utf-8 -*-
# vispy: testskip
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
"""
Compare an optimal plot grid implementation to the same functionality
provided by scenegraph.

Use --vispy-cprofile to see an overview of time spent in all functions.
Use util.profiler and --vispy-profile=ClassName.method_name for more directed
profiling measurements.
"""
from __future__ import division
import numpy as np

from vispy import gloo, app, scene, visuals
from vispy.util.profiler import Profiler


class GridCanvas(app.Canvas):
    def __init__(self, cells, **kwargs):
        m, n = (10, 10)
        self.grid_size = (m, n)
        self.cells = cells
        super(GridCanvas, self).__init__(keys='interactive',
                                         show=True, **kwargs)

    def on_initialize(self, event):
        self.context.set_state(clear_color='black', blend=True,
                               blend_func=('src_alpha', 'one_minus_src_alpha'))

    def on_mouse_move(self, event):
        if event.is_dragging and not event.modifiers:
            dx = (event.pos - event.last_event.pos) * [1, -1]
            i, j = event.press_event.pos / self.size
            m, n = len(self.cells), len(self.cells[0])
            cell = self.cells[int(i*m)][n - 1 - int(j*n)]
            if event.press_event.button == 1:
                offset = (np.array(cell.offset) + 
                          (dx / (np.array(self.size) / [m, n])) *  
                          (2 / np.array(cell.scale)))
                cell.set_transform(offset, cell.scale)

            else:
                cell.set_transform(cell.offset, cell.scale * 1.05 ** dx)
            self.update()

    def on_draw(self, event):
        prof = Profiler()  # noqa
        self.context.clear()
        M = len(self.cells)
        N = len(self.cells[0])
        w, h = self.size
        for i in range(M):
            for j in range(N):
                self.context.set_viewport(w*i/M, h*j/N, w/M, h/N)
                self.cells[i][j].draw()


vert = """
attribute vec2 pos;
uniform vec2 offset;
uniform vec2 scale;

void main() {
    gl_Position = vec4((pos + offset) * scale, 0, 1);
}
"""

frag = """
void main() {
    gl_FragColor = vec4(1, 1, 1, 0.5);
}
"""


class Line(object):
    def __init__(self, data, offset, scale):
        self.data = gloo.VertexBuffer(data)
        self.program = gloo.Program(vert, frag)
        self.program['pos'] = self.data
        self.set_transform(offset, scale)

    def set_transform(self, offset, scale):
        self.offset = offset
        self.scale = scale
        self.program['offset'] = self.offset
        self.program['scale'] = self.scale

    def draw(self):
        self.program.draw('line_strip')


scales = np.array((1.9 / 100., 2. / 10.))


class VisualCanvas(app.Canvas):
    def __init__(self, vis, **kwargs):
        super(VisualCanvas, self).__init__(keys='interactive',
                                           show=True, **kwargs)
        m, n = (10, 10)
        self.grid_size = (m, n)
        self.visuals = vis

    def on_initialize(self, event):
        self.context.set_state(clear_color='black', blend=True,
                               blend_func=('src_alpha', 'one_minus_src_alpha'))

    def on_mouse_move(self, event):
        if event.is_dragging and not event.modifiers:
            dx = np.array(event.pos - event.last_event.pos)
            x, y = event.press_event.pos / self.size
            m, n = self.grid_size
            i, j = int(x*m), n - 1 - int(y*n)
            v = self.visuals[i][j]
            tr = v.transform
            if event.press_event.button == 1:
                tr.translate = np.array(tr.translate)[:2] + \
                    dx * scales * (1, -1)

            else:
                tr.scale = tr.scale[:2] * 1.05 ** (dx * (1, -1))
            self.update()

    def on_draw(self, event):
        prof = Profiler()  # noqa
        self.context.clear()
        M, N = self.grid_size
        w, h = self.size
        for i in range(M):
            for j in range(N):
                self.context.set_viewport(w*i/M, h*j/N, w/M, h/N)
                self.visuals[i][j].draw()


if __name__ == '__main__':
    M, N = (10, 10)

    data = np.empty((10000, 2), dtype=np.float32)
    data[:, 0] = np.linspace(0, 100, data.shape[0])
    data[:, 1] = np.random.normal(size=data.shape[0])

    # Optimized version
    cells = []
    for i in range(M):
        row = []
        cells.append(row)
        for j in range(N):
            row.append(Line(data, offset=(-50, 0), scale=scales))

    gcanvas = GridCanvas(cells, position=(400, 300), size=(800, 600),
                         title="GridCanvas")

    # Visual version
    vlines = []
    for i in range(M):
        row = []
        vlines.append(row)
        for j in range(N):
            v = visuals.LineVisual(pos=data, color='w', method='gl')
            v.transform = visuals.transforms.STTransform(
                translate=(-1, 0), scale=scales)
            row.append(v)

    vcanvas = VisualCanvas(vlines, position=(400, 300), size=(800, 600), 
                           title="VisualCanvas")

    # Scenegraph version
    scanvas = scene.SceneCanvas(show=True, keys='interactive', 
                                title="SceneCanvas")

    scanvas.size = 800, 600
    grid = scanvas.central_widget.add_grid(margin=0)

    lines = []
    for i in range(10):
        lines.append([])
        for j in range(10):
            vb = grid.add_view(camera='panzoom', row=i, col=j)
            vb.camera.set_range([0, 100], [-5, 5], margin=0)
            line = scene.visuals.Line(pos=data, color='w', method='gl')
            vb.add(line)
    scanvas.show()

    import sys
    if sys.flags.interactive != 1:
        app.run()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
"""
Tutorial: Creating Visuals
--------------------------

This tutorial is intended to guide developers who are interested in creating 
new subclasses of Visual. In most cases, this will not be necessary because
vispy's base library of visuals will be sufficient to create complex scenes as
needed. However, there are cases where a particular visual effect is desired 
that is not supported in the base library, or when a custom visual is needed to
optimize performance for a specific use case.

The purpose of a Visual is to encapsulate a single drawable object. This
drawable can be as simple or complex as desired. Some of the simplest visuals 
draw points, lines, or triangles, whereas more complex visuals invove multiple
drawing stages or make use of sub-visuals to construct larger objects.

In this example we will create a very simple Visual that draws a rectangle.
Visuals are defined by:

1. Creating a subclass of vispy.visuals.Visual that specifies the GLSL code
   and buffer objects to use.
2. Defining a _prepare_transforms() method that will be called whenever the
   user (or scenegraph) assigns a new set of transforms to the visual.


"""
from vispy import app, gloo, visuals, scene
import numpy as np

# Define a simple vertex shader. We use $template variables as placeholders for
# code that will be inserted later on. In this example, $position will become
# an attribute, and $transform will become a function. Important: using
# $transform in this way ensures that users of this visual will be able to 
# apply arbitrary transformations to it.
vertex_shader = """
void main() {
   gl_Position = $transform(vec4($position, 0, 1));
}
"""

# Very simple fragment shader. Again we use a template variable "$color", which
# allows us to decide later how the color should be defined (in this case, we
# will just use a uniform red color).
fragment_shader = """
void main() {
  gl_FragColor = $color;
}
"""


# Start the new Visual class. 
# By convention, all Visual subclass names end in 'Visual'.
# (Custom visuals may ignore this convention, but for visuals that are built 
# in to vispy, this is required to ensure that the VisualNode subclasses are 
# generated correctly.)
class MyRectVisual(visuals.Visual):
    """Visual that draws a red rectangle.

    Parameters
    ----------
    x : float
        x coordinate of rectangle origin
    y : float
        y coordinate of rectangle origin
    w : float
        width of rectangle
    h : float
        height of rectangle

    All parameters are specified in the local (arbitrary) coordinate system of
    the visual. How this coordinate system translates to the canvas will 
    depend on the transformation functions used during drawing.
    """

    # There are no constraints on the signature of the __init__ method; use
    # whatever makes the most sense for your visual.
    def __init__(self, x, y, w, h):
        # Initialize the visual with a vertex shader and fragment shader
        visuals.Visual.__init__(self, vertex_shader, fragment_shader)

        # vertices for two triangles forming a rectangle
        self.vbo = gloo.VertexBuffer(np.array([
            [x, y], [x+w, y], [x+w, y+h],
            [x, y], [x+w, y+h], [x, y+h]
        ], dtype=np.float32))

        # Assign values to the $position and $color template variables in 
        # the shaders. ModularProgram automatically handles generating the 
        # necessary attribute and uniform declarations with unique variable
        # names.
        self.shared_program.vert['position'] = self.vbo
        self.shared_program.frag['color'] = (1, 0, 0, 1)
        self._draw_mode = 'triangles'

    def _prepare_transforms(self, view):
        # This method is called when the user or the scenegraph has assigned
        # new transforms to this visual (ignore the *view* argument for now;
        # we'll get to that later). This method is thus responsible for
        # connecting the proper transform functions to the shader program.

        # The most common approach here is to simply take the complete
        # transformation from visual coordinates to render coordinates. Later
        # tutorials detail more complex transform handling.
        view.view_program.vert['transform'] = view.get_transform()


# At this point the visual is ready to use, but it takes some extra effort to
# set up a Canvas and TransformSystem for drawing (the examples in 
# examples/basics/visuals/ all follow this approach). 
# 
# An easier approach is to make the visual usable in a scenegraph, in which 
# case the canvas will take care of drawing the visual and setting up the 
# TransformSystem for us.
# 
# To be able to use our new Visual in a scenegraph, it needs to be
# a subclass of scene.Node. In vispy we achieve this by creating a parallel
# set of classes that inherit from both Node and each Visual subclass.
# This can be done automatically using scene.visuals.create_visual_node():
MyRect = scene.visuals.create_visual_node(MyRectVisual)

# By convention, these classes have the same name as the Visual they inherit 
# from, but without the 'Visual' suffix.

# The auto-generated class MyRect is basically equivalent to::
# 
#     class MyRect(MyRectVisual, scene.Node):
#        def __init__(self, *args, **kwds):
#            parent = kwds.pop('parent', None)
#            name = kwds.pop('name', None)
#            MyRectVisual.__init__(self, *args, **kwds)
#            Node.__init__(self, parent=parent, name=name)
#         


# Finally we will test the visual by displaying in a scene.

# Create a canvas to display our visual
canvas = scene.SceneCanvas(keys='interactive', show=True)

# Create two instances of MyRect, each using canvas.scene as their parent
rects = [MyRect(100, 100, 200, 300, parent=canvas.scene),
         MyRect(500, 100, 200, 300, parent=canvas.scene)]

# To test that the user-specified transforms work correctly, I'll rotate
# one rectangle slightly.
tr = visuals.transforms.MatrixTransform()
tr.rotate(5, (0, 0, 1))
rects[1].transform = tr

# ..and optionally start the event loop
if __name__ == '__main__':
    import sys
    if sys.flags.interactive != 1:
        app.run()



# -*- coding: utf-8 -*-
# vispy: gallery 30
# -----------------------------------------------------------------------------
# Copyright (c) Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
""" Display Windbarbs.
"""

import itertools

import numpy as np
from vispy import app
from vispy.visuals import WindbarbVisual
from vispy.visuals.transforms import NullTransform


class Canvas(app.Canvas):
    def __init__(self):
        app.Canvas.__init__(self, title="Windbarb plot", keys="interactive",
                            size=(830, 430))

        self.windbarb_length = 25
        self.grid_spacing = 50

        self.grid_coords = None
        self.line_vertices = None
        self.last_mouse = (0, 0)

        self.generate_grid()

        direction_vectors = (self.grid_coords - self.last_mouse).astype(
            np.float32)
        direction_vectors[:] /= 10
        direction_vectors[:, 1] *= -1

        self.visual = WindbarbVisual(pos=self.grid_coords,
                                     wind=direction_vectors,
                                     trig=False,
                                     edge_color='black',
                                     face_color='black',
                                     size=self.windbarb_length)

        self.visual.events.update.connect(lambda evt: self.update())
        self.visual.transform = NullTransform()

        self.show()

    def generate_grid(self):
        n = self.grid_spacing / 2.
        num_cols = int(self.physical_size[0] / n * 2)
        num_rows = int(self.physical_size[1] / n * 2)
        coords = []

        # Generate grid
        for i, j in itertools.product(range(num_rows), range(num_cols)):
            x = n + (n * 2 * j)
            y = n + (n * 2 * i)

            coords.append((x, y))

        self.grid_coords = np.array(coords)

    def on_resize(self, event):
        self.generate_grid()
        self.rotate_arrows(np.array(self.last_mouse))

        vp = (0, 0, self.physical_size[0], self.physical_size[1])
        self.context.set_viewport(*vp)
        self.visual.transforms.configure(canvas=self, viewport=vp)

    def rotate_arrows(self, point_towards):
        direction_vectors = (self.grid_coords - point_towards).astype(
            np.float32)
        direction_vectors[:] /= 10
        direction_vectors[:, 1] *= -1

        self.visual.set_data(
            pos=self.grid_coords,
            wind=direction_vectors,
            size=self.windbarb_length,
        )

    def on_mouse_move(self, event):
        self.last_mouse = event.pos
        self.rotate_arrows(np.array(event.pos))

    def on_draw(self, event):
        self.context.clear(color='white')
        self.visual.draw()


if __name__ == '__main__':
    canvas = Canvas()
    canvas.measure_fps()
    app.run()
